from .gaussian import (
    NumPyGaussianSampler as NumPyGaussianSampler,
    JaxGaussianSampler as JaxGaussianSampler,
)
from .halton import (
    NumPyHaltonSplineSampler as NumPyHaltonSplineSampler,
    JaxHaltonSplineSampler as JaxHaltonSplineSampler,
)
from .factory import sampler as sampler



from typing import Final, Any

from trajax.types import (
    NumPyState,
    NumPyStateSequence,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPyCosts,
    NumPyCostFunction,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxCosts,
    JaxCostFunction,
    State as State,
    StateSequence as StateSequence,
    StateBatch as StateBatch,
    ControlInputSequence as ControlInputSequence,
    ControlInputBatch as ControlInputBatch,
    Costs as Costs,
    CostFunction as CostFunction,
    BICYCLE_D_X,
    BICYCLE_D_U,
    BicycleD_x,
    BicycleD_u,
    BicycleState,
    BicycleStateSequence,
    BicycleStateBatch,
    BicyclePositions,
    BicycleControlInputSequence,
    BicycleControlInputBatch,
    NumPyPathParameters,
    NumPyReferencePoints,
    NumPyPositions,
    NumPyHeadings,
    NumPyLateralPositions,
    NumPyLongitudinalPositions,
    NumPyNormals,
    NumPyRisk,
    JaxPathParameters,
    JaxReferencePoints,
    JaxPositions,
    JaxHeadings,
    JaxLateralPositions,
    JaxLongitudinalPositions,
    JaxNormals,
    JaxRisk,
    D_o as D_o_,
    D_O as D_O_,
    NumPyPathParameterExtractor,
    NumPyPathVelocityExtractor,
    NumPyPositionExtractor,
    NumPyDistanceExtractor,
    NumPyRiskMetric,
    JaxPathParameterExtractor,
    JaxPathVelocityExtractor,
    JaxPositionExtractor,
    JaxDistanceExtractor,
    JaxRiskMetric,
    Error as Error,  # NOTE: Aliased to workaround ruff bug.
    Risk as Risk,
    RiskMetric as RiskMetric,
    ContouringCost as ContouringCost,
    NumPyObstacleStateProvider,
    NumPyObstaclePositionExtractor,
    JaxObstacleStateProvider,
    JaxObstaclePositionExtractor,
    AugmentedState,
    AugmentedStateSequence,
    AugmentedStateBatch,
    AugmentedControlInputSequence,
    AugmentedControlInputBatch,
    NumPyInitialPositionCovariance,
    NumPyInitialVelocityCovariance,
    NumPyInitialCovarianceProvider,
    NumPyPositionCovariance,
    JaxInitialPositionCovariance,
    JaxInitialVelocityCovariance,
    JaxInitialCovarianceProvider,
    JaxPositionCovariance,
    NumPyBoundaryDistance,
    NumPyBoundaryDistanceExtractor,
    JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor,
)
from trajax.models import (
    NumPyBicycleState,
    NumPyBicycleStateSequence,
    NumPyBicycleStateBatch,
    NumPyBicyclePositions,
    NumPyBicycleControlInputSequence,
    NumPyBicycleControlInputBatch,
    NumPyBicycleObstacleStateSequences,
    NumPyIntegratorObstacleStateSequences,
    JaxBicycleState,
    JaxBicycleStateSequence,
    JaxBicycleStateBatch,
    JaxBicyclePositions,
    JaxBicycleControlInputSequence,
    JaxBicycleControlInputBatch,
    JaxBicycleObstacleStateSequences,
    JaxIntegratorObstacleStateSequences,
)
from trajax.costs import (
    NumPyContouringCost,
    JaxContouringCost,
    NumPyDistance,
    JaxDistance,
)
from trajax.obstacles import (
    NumPySampledObstacleStates,
    NumPyObstacleIds,
    NumPyObstacleStates,
    NumPyObstacleStatesForTimeStep,
    NumPyObstacle2dPositions,
    NumPyObstacle2dPositionsForTimeStep,
    NumPyObstacleStatesRunningHistory,
    JaxSampledObstacleStates,
    JaxObstacleIds,
    JaxObstacleStates,
    JaxObstacleStatesForTimeStep,
    JaxObstacle2dPositions,
    JaxObstacle2dPositionsForTimeStep,
    JaxObstacleStatesRunningHistory,
)
from trajax.states import (
    NumPySimpleState,
    NumPySimpleStateSequence,
    NumPySimpleStateBatch,
    NumPySimpleControlInputSequence,
    NumPySimpleControlInputBatch,
    NumPySimpleCosts,
    JaxSimpleState,
    JaxSimpleStateSequence,
    JaxSimpleStateBatch,
    JaxSimpleControlInputSequence,
    JaxSimpleControlInputBatch,
    JaxSimpleCosts,
    NumPyAugmentedState,
    NumPyAugmentedStateSequence,
    NumPyAugmentedStateBatch,
    NumPyAugmentedControlInputSequence,
    NumPyAugmentedControlInputBatch,
    JaxAugmentedState,
    JaxAugmentedStateSequence,
    JaxAugmentedStateBatch,
    JaxAugmentedControlInputSequence,
    JaxAugmentedControlInputBatch,
)


class types:
    type State[D_x: int = Any] = State[D_x]
    type StateBatch[T: int = Any, D_x: int = Any, M: int = Any] = StateBatch[T, D_x, M]
    type ControlInputSequence[T: int = Any, D_u: int = Any] = ControlInputSequence[
        T, D_u
    ]
    type ControlInputBatch[T: int = Any, D_u: int = Any, M: int = Any] = (
        ControlInputBatch[T, D_u, M]
    )
    type Costs[T: int = Any, M: int = Any] = Costs[T, M]
    type CostFunction[I, S, C] = CostFunction[I, S, C]
    type Error[T: int = Any, M: int = Any] = Error[T, M]
    type Risk[T: int = Any, M: int = Any] = Risk[T, M]

    type ContouringCost[I, S, E] = ContouringCost[I, S, E]
    type RiskMetric[CF, SB, OS, S, R] = RiskMetric[CF, SB, OS, S, R]

    class obstacle:
        type D_o = D_o_

        D_O: Final = D_O_

    class bicycle:
        type D_x = BicycleD_x
        type D_u = BicycleD_u
        type State = BicycleState
        type StateSequence = BicycleStateSequence
        type StateBatch[T: int = Any, M: int = Any] = BicycleStateBatch[T, M]
        type Positions[T: int = Any, M: int = Any] = BicyclePositions[T, M]
        type ControlInputSequence[T: int = Any] = BicycleControlInputSequence[T]
        type ControlInputBatch[T: int = Any, M: int = Any] = BicycleControlInputBatch[
            T, M
        ]

        D_X: Final = BICYCLE_D_X
        D_U: Final = BICYCLE_D_U

    class augmented:
        type State[P, V] = AugmentedState[P, V]
        type StateSequence[P, V] = AugmentedStateSequence[P, V]
        type StateBatch[P, V] = AugmentedStateBatch[P, V]
        type ControlInputSequence[P, V] = AugmentedControlInputSequence[P, V]
        type ControlInputBatch[P, V] = AugmentedControlInputBatch[P, V]

        state: Final = AugmentedState
        state_sequence: Final = AugmentedStateSequence
        state_batch: Final = AugmentedStateBatch
        control_input_sequence: Final = AugmentedControlInputSequence
        control_input_batch: Final = AugmentedControlInputBatch

    class numpy:
        type State[D_x: int = Any] = NumPyState[D_x]
        type StateSequence[T: int = Any, D_x: int = Any] = NumPyStateSequence[T, D_x]
        type StateBatch[T: int = Any, D_x: int = Any, M: int = Any] = NumPyStateBatch[
            T, D_x, M
        ]
        type ControlInputSequence[T: int = Any, D_u: int = Any] = (
            NumPyControlInputSequence[T, D_u]
        )
        type ControlInputBatch[T: int = Any, D_u: int = Any, M: int = Any] = (
            NumPyControlInputBatch[T, D_u, M]
        )
        type Costs[T: int = Any, M: int = Any] = NumPyCosts[T, M]
        type PathParameters[T: int = Any, M: int = Any] = NumPyPathParameters[T, M]
        type ReferencePoints[T: int = Any, M: int = Any] = NumPyReferencePoints[T, M]
        type Positions[T: int = Any, M: int = Any] = NumPyPositions[T, M]
        type Headings[T: int = Any, M: int = Any] = NumPyHeadings[T, M]
        type LateralPositions[T: int = Any, M: int = Any] = NumPyLateralPositions[T, M]
        type LongitudinalPositions[T: int = Any, M: int = Any] = (
            NumPyLongitudinalPositions[T, M]
        )
        type ObstacleIds[K: int = Any] = NumPyObstacleIds[K]
        type ObstacleStates[T: int = Any, K: int = Any] = NumPyObstacleStates[T, K]
        type SampledObstacleStates[T: int = Any, K: int = Any, N: int = Any] = (
            NumPySampledObstacleStates[T, K, N]
        )
        type ObstacleStatesForTimeStep[K: int = Any] = NumPyObstacleStatesForTimeStep[K]
        type Obstacle2dPositions[T: int = Any, K: int = Any] = NumPyObstacle2dPositions[
            T, K
        ]
        type Obstacle2dPositionsForTimeStep[K: int = Any] = (
            NumPyObstacle2dPositionsForTimeStep[K]
        )
        type ObstacleStatesRunningHistory[T: int = Any, K: int = Any] = (
            NumPyObstacleStatesRunningHistory[T, K]
        )
        type Distance[T: int = Any, V: int = Any, M: int = Any, N: int = Any] = (
            NumPyDistance[T, V, M, N]
        )
        type BoundaryDistance[T: int = Any, M: int = Any] = NumPyBoundaryDistance[T, M]
        type Risk[T: int = Any, M: int = Any] = NumPyRisk[T, M]
        type InitialPositionCovariance[K: int = Any] = NumPyInitialPositionCovariance[K]
        type InitialVelocityCovariance[K: int = Any] = NumPyInitialVelocityCovariance[K]
        type PositionCovariance[T: int = Any, K: int = Any] = NumPyPositionCovariance[
            T, K
        ]

        type CostFunction[
            T: int = Any,
            D_u: int = Any,
            D_x: int = Any,
            M: int = Any,
        ] = NumPyCostFunction[
            NumPyControlInputBatch[T, D_u, M],
            NumPyStateBatch[T, D_x, M],
            NumPyCosts[T, M],
        ]
        type PathParameterExtractor[S] = NumPyPathParameterExtractor[S]
        type PathVelocityExtractor[I] = NumPyPathVelocityExtractor[I]
        type PositionExtractor[S] = NumPyPositionExtractor[S]
        type DistanceExtractor[SB, SOS, D] = NumPyDistanceExtractor[SB, SOS, D]
        type BoundaryDistanceExtractor[SB, D] = NumPyBoundaryDistanceExtractor[SB, D]
        type RiskMetric[SB, OS, SOS] = NumPyRiskMetric[SB, OS, SOS]
        type ContouringCost[S] = NumPyContouringCost[S]
        type ObstacleStateProvider[O] = NumPyObstacleStateProvider[O]
        type ObstaclePositionExtractor[OTS, O, PTS, P] = NumPyObstaclePositionExtractor[
            OTS, O, PTS, P
        ]
        type InitialCovarianceProvider[S] = NumPyInitialCovarianceProvider[S]

        path_parameters: Final = NumPyPathParameters
        reference_points: Final = NumPyReferencePoints.create
        positions: Final = NumPyPositions.create
        headings: Final = NumPyHeadings.create
        lateral_positions: Final = NumPyLateralPositions.create
        longitudinal_positions: Final = NumPyLongitudinalPositions.create
        normals: Final = NumPyNormals.create
        distance: Final = NumPyDistance
        boundary_distance: Final = NumPyBoundaryDistance
        obstacle_ids: Final = NumPyObstacleIds
        obstacle_states: Final = NumPyObstacleStates
        obstacle_states_for_time_step: Final = NumPyObstacleStatesForTimeStep
        obstacle_states_running_history: Final = NumPyObstacleStatesRunningHistory

        class simple:
            type State[D_x: int = Any] = NumPySimpleState[D_x]
            type StateSequence[T: int = Any, D_x: int = Any] = NumPySimpleStateSequence[
                T, D_x
            ]
            type StateBatch[T: int = Any, D_x: int = Any, M: int = Any] = (
                NumPySimpleStateBatch[T, D_x, M]
            )
            type ControlInputSequence[T: int = Any, D_u: int = Any] = (
                NumPySimpleControlInputSequence[T, D_u]
            )
            type ControlInputBatch[T: int = Any, D_u: int = Any, M: int = Any] = (
                NumPySimpleControlInputBatch[T, D_u, M]
            )
            type Costs[T: int = Any, M: int = Any] = NumPySimpleCosts[T, M]

            state: Final = NumPySimpleState
            state_sequence: Final = NumPySimpleStateSequence
            state_batch: Final = NumPySimpleStateBatch
            control_input_sequence: Final = NumPySimpleControlInputSequence
            control_input_batch: Final = NumPySimpleControlInputBatch
            costs: Final = NumPySimpleCosts

        class integrator:
            type ObstacleStateSequences[T: int = Any, D_o: int = Any, K: int = Any] = (
                NumPyIntegratorObstacleStateSequences[T, D_o, K]
            )

        class bicycle:
            type State = NumPyBicycleState
            type StateSequence[T: int = Any] = NumPyBicycleStateSequence[T]
            type StateBatch[T: int = Any, M: int = Any] = NumPyBicycleStateBatch[T, M]
            type Positions[T: int = Any, M: int = Any] = NumPyBicyclePositions[T, M]
            type ControlInputSequence[T: int = Any] = NumPyBicycleControlInputSequence[
                T
            ]
            type ControlInputBatch[T: int = Any, M: int = Any] = (
                NumPyBicycleControlInputBatch[T, M]
            )
            type ObstacleStateSequences[T: int = Any, K: int = Any] = (
                NumPyBicycleObstacleStateSequences[T, K]
            )

            state: Final = NumPyBicycleState
            state_sequence: Final = NumPyBicycleStateSequence
            state_batch: Final = NumPyBicycleStateBatch
            positions: Final = NumPyBicyclePositions
            control_input_sequence: Final = NumPyBicycleControlInputSequence
            control_input_batch: Final = NumPyBicycleControlInputBatch

        class augmented:
            type State[P: NumPyState, V: NumPyState] = NumPyAugmentedState[P, V]
            type StateSequence[P: NumPyStateSequence, V: NumPyStateSequence] = (
                NumPyAugmentedStateSequence[P, V]
            )
            type StateBatch[P: NumPyStateBatch, V: NumPyStateBatch] = (
                NumPyAugmentedStateBatch[P, V]
            )
            type ControlInputSequence[
                P: NumPyControlInputSequence,
                V: NumPyControlInputSequence,
            ] = NumPyAugmentedControlInputSequence[P, V]
            type ControlInputBatch[
                P: NumPyControlInputBatch,
                V: NumPyControlInputBatch,
            ] = NumPyAugmentedControlInputBatch[P, V]

            state: Final = NumPyAugmentedState
            state_sequence: Final = NumPyAugmentedStateSequence
            state_batch: Final = NumPyAugmentedStateBatch
            control_input_sequence: Final = NumPyAugmentedControlInputSequence
            control_input_batch: Final = NumPyAugmentedControlInputBatch

    class jax:
        type State[D_x: int = Any] = JaxState[D_x]
        type StateSequence[T: int = Any, D_x: int = Any] = JaxStateSequence[T, D_x]
        type StateBatch[T: int = Any, D_x: int = Any, M: int = Any] = JaxStateBatch[
            T, D_x, M
        ]
        type ControlInputSequence[T: int = Any, D_u: int = Any] = (
            JaxControlInputSequence[T, D_u]
        )
        type ControlInputBatch[T: int = Any, D_u: int = Any, M: int = Any] = (
            JaxControlInputBatch[T, D_u, M]
        )
        type Costs[T: int = Any, M: int = Any] = JaxCosts[T, M]
        type PathParameters[T: int = Any, M: int = Any] = JaxPathParameters[T, M]
        type ReferencePoints[T: int = Any, M: int = Any] = JaxReferencePoints[T, M]
        type Positions[T: int = Any, M: int = Any] = JaxPositions[T, M]
        type Headings[T: int = Any, M: int = Any] = JaxHeadings[T, M]
        type LateralPositions[T: int = Any, M: int = Any] = JaxLateralPositions[T, M]
        type LongitudinalPositions[T: int = Any, M: int = Any] = (
            JaxLongitudinalPositions[T, M]
        )
        type ObstacleIds[K: int = Any] = JaxObstacleIds[K]
        type ObstacleStates[T: int = Any, K: int = Any] = JaxObstacleStates[T, K]
        type SampledObstacleStates[T: int = Any, K: int = Any, N: int = Any] = (
            JaxSampledObstacleStates[T, K, N]
        )
        type ObstacleStatesForTimeStep[K: int = Any] = JaxObstacleStatesForTimeStep[K]
        type Obstacle2dPositions[T: int = Any, K: int = Any] = JaxObstacle2dPositions[
            T, K
        ]
        type Obstacle2dPositionsForTimeStep[K: int = Any] = (
            JaxObstacle2dPositionsForTimeStep[K]
        )
        type ObstacleStatesRunningHistory[T: int = Any, K: int = Any] = (
            JaxObstacleStatesRunningHistory[T, K]
        )
        type Distance[T: int = Any, V: int = Any, M: int = Any, N: int = Any] = (
            JaxDistance[T, V, M, N]
        )
        type BoundaryDistance[T: int = Any, M: int = Any] = JaxBoundaryDistance[T, M]
        type Risk[T: int = Any, M: int = Any] = JaxRisk[T, M]
        type InitialPositionCovariance[K: int = Any] = JaxInitialPositionCovariance[K]
        type InitialVelocityCovariance[K: int = Any] = JaxInitialVelocityCovariance[K]
        type PositionCovariance[T: int = Any, K: int = Any] = JaxPositionCovariance[
            T, K
        ]

        type CostFunction[
            T: int = Any,
            D_u: int = Any,
            D_x: int = Any,
            M: int = Any,
        ] = JaxCostFunction[
            JaxControlInputBatch[T, D_u, M], JaxStateBatch[T, D_x, M], JaxCosts[T, M]
        ]
        type PathParameterExtractor[S] = JaxPathParameterExtractor[S]
        type PathVelocityExtractor[I] = JaxPathVelocityExtractor[I]
        type PositionExtractor[S] = JaxPositionExtractor[S]
        type DistanceExtractor[SB, SOS, D] = JaxDistanceExtractor[SB, SOS, D]
        type BoundaryDistanceExtractor[SB, D] = JaxBoundaryDistanceExtractor[SB, D]
        type RiskMetric[SB, OS, SOS] = JaxRiskMetric[SB, OS, SOS]
        type ContouringCost[S] = JaxContouringCost[S]
        type ObstacleStateProvider[O] = JaxObstacleStateProvider[O]
        type ObstaclePositionExtractor[OTS, O, PTS, P] = JaxObstaclePositionExtractor[
            OTS, O, PTS, P
        ]
        type InitialCovarianceProvider[S] = JaxInitialCovarianceProvider[S]

        path_parameters: Final = JaxPathParameters.create
        reference_points: Final = JaxReferencePoints.create
        positions: Final = JaxPositions.create
        headings: Final = JaxHeadings.create
        lateral_positions: Final = JaxLateralPositions.create
        longitudinal_positions: Final = JaxLongitudinalPositions.create
        normals: Final = JaxNormals.create
        distance: Final = JaxDistance
        boundary_distance: Final = JaxBoundaryDistance
        obstacle_ids: Final = JaxObstacleIds
        obstacle_states: Final = JaxObstacleStates
        obstacle_states_for_time_step: Final = JaxObstacleStatesForTimeStep
        obstacle_states_running_history: Final = JaxObstacleStatesRunningHistory

        class simple:
            type State[D_x: int = Any] = JaxSimpleState[D_x]
            type StateSequence[T: int = Any, D_x: int = Any] = JaxSimpleStateSequence[
                T, D_x
            ]
            type StateBatch[T: int = Any, D_x: int = Any, M: int = Any] = (
                JaxSimpleStateBatch[T, D_x, M]
            )
            type ControlInputSequence[T: int = Any, D_u: int = Any] = (
                JaxSimpleControlInputSequence[T, D_u]
            )
            type ControlInputBatch[T: int = Any, D_u: int = Any, M: int = Any] = (
                JaxSimpleControlInputBatch[T, D_u, M]
            )
            type Costs[T: int = Any, M: int = Any] = JaxSimpleCosts[T, M]

            state: Final = JaxSimpleState
            state_sequence: Final = JaxSimpleStateSequence
            state_batch: Final = JaxSimpleStateBatch
            control_input_sequence: Final = JaxSimpleControlInputSequence
            control_input_batch: Final = JaxSimpleControlInputBatch
            costs: Final = JaxSimpleCosts

        class integrator:
            type ObstacleStateSequences[T: int = Any, D_o: int = Any, K: int = Any] = (
                JaxIntegratorObstacleStateSequences[T, D_o, K]
            )

        class bicycle:
            type State = JaxBicycleState
            type StateSequence[T: int = Any] = JaxBicycleStateSequence[T]
            type StateBatch[T: int = Any, M: int = Any] = JaxBicycleStateBatch[T, M]
            type Positions[T: int = Any, M: int = Any] = JaxBicyclePositions[T, M]
            type ControlInputSequence[T: int = Any] = JaxBicycleControlInputSequence[T]
            type ControlInputBatch[T: int = Any, M: int = Any] = (
                JaxBicycleControlInputBatch[T, M]
            )
            type ObstacleStateSequences[T: int = Any, K: int = Any] = (
                JaxBicycleObstacleStateSequences[T, K]
            )

            state: Final = JaxBicycleState
            state_sequence: Final = JaxBicycleStateSequence
            state_batch: Final = JaxBicycleStateBatch
            positions: Final = JaxBicyclePositions
            control_input_sequence: Final = JaxBicycleControlInputSequence
            control_input_batch: Final = JaxBicycleControlInputBatch

        class augmented:
            type State[P: JaxState, V: JaxState] = JaxAugmentedState[P, V]
            type StateSequence[P: JaxStateSequence, V: JaxStateSequence] = (
                JaxAugmentedStateSequence[P, V]
            )
            type StateBatch[P: JaxStateBatch, V: JaxStateBatch] = (
                JaxAugmentedStateBatch[P, V]
            )
            type ControlInputSequence[
                P: JaxControlInputSequence,
                V: JaxControlInputSequence,
            ] = JaxAugmentedControlInputSequence[P, V]
            type ControlInputBatch[P: JaxControlInputBatch, V: JaxControlInputBatch] = (
                JaxAugmentedControlInputBatch[P, V]
            )

            state: Final = JaxAugmentedState
            state_sequence: Final = JaxAugmentedStateSequence
            state_batch: Final = JaxAugmentedStateBatch
            control_input_sequence: Final = JaxAugmentedControlInputSequence
            control_input_batch: Final = JaxAugmentedControlInputBatch


class classes:
    State: Final = State
    StateSequence: Final = StateSequence
    StateBatch: Final = StateBatch
    ControlInputSequence: Final = ControlInputSequence
    ControlInputBatch: Final = ControlInputBatch
    Costs: Final = Costs
    CostFunction: Final = CostFunction
    Error: Final = Error
    ContouringCost: Final = ContouringCost

    class bicycle:
        State: Final = BicycleState
        StateSequence: Final = BicycleStateSequence
        StateBatch: Final = BicycleStateBatch
        Positions: Final = BicyclePositions
        ControlInputSequence: Final = BicycleControlInputSequence
        ControlInputBatch: Final = BicycleControlInputBatch

    class augmented:
        State: Final = AugmentedState
        StateSequence: Final = AugmentedStateSequence
        StateBatch: Final = AugmentedStateBatch
        ControlInputSequence: Final = AugmentedControlInputSequence
        ControlInputBatch: Final = AugmentedControlInputBatch

    class numpy:
        State: Final = NumPyState
        StateSequence: Final = NumPyStateSequence
        StateBatch: Final = NumPyStateBatch
        ControlInputSequence: Final = NumPyControlInputSequence
        ControlInputBatch: Final = NumPyControlInputBatch
        Costs: Final = NumPyCosts
        PathParameters: Final = NumPyPathParameters
        ReferencePoints: Final = NumPyReferencePoints
        Positions: Final = NumPyPositions
        Headings: Final = NumPyHeadings
        ObstacleStates: Final = NumPyObstacleStates
        SampledObstacleStates: Final = NumPySampledObstacleStates
        Distance: Final = NumPyDistance
        CostFunction: Final = NumPyCostFunction
        PathParameterExtractor: Final = NumPyPathParameterExtractor
        PathVelocityExtractor: Final = NumPyPathVelocityExtractor
        PositionExtractor: Final = NumPyPositionExtractor
        ContouringCost: Final = NumPyContouringCost
        ObstacleStateProvider: Final = NumPyObstacleStateProvider

        class simple:
            State: Final = NumPySimpleState
            StateSequence: Final = NumPySimpleStateSequence
            StateBatch: Final = NumPySimpleStateBatch
            ControlInputSequence: Final = NumPySimpleControlInputSequence
            ControlInputBatch: Final = NumPySimpleControlInputBatch
            Costs: Final = NumPySimpleCosts

        class integrator:
            ObstacleStateSequences: Final = NumPyIntegratorObstacleStateSequences

        class bicycle:
            State: Final = NumPyBicycleState
            StateSequence: Final = NumPyBicycleStateSequence
            StateBatch: Final = NumPyBicycleStateBatch
            Positions: Final = NumPyBicyclePositions
            ControlInputSequence: Final = NumPyBicycleControlInputSequence
            ControlInputBatch: Final = NumPyBicycleControlInputBatch
            ObstacleStateSequences: Final = NumPyBicycleObstacleStateSequences

        class augmented:
            State: Final = NumPyAugmentedState
            StateSequence: Final = NumPyAugmentedStateSequence
            StateBatch: Final = NumPyAugmentedStateBatch
            ControlInputSequence: Final = NumPyAugmentedControlInputSequence
            ControlInputBatch: Final = NumPyAugmentedControlInputBatch

    class jax:
        State: Final = JaxState
        StateSequence: Final = JaxStateSequence
        StateBatch: Final = JaxStateBatch
        ControlInputSequence: Final = JaxControlInputSequence
        ControlInputBatch: Final = JaxControlInputBatch
        Costs: Final = JaxCosts
        PathParameters: Final = JaxPathParameters
        ReferencePoints: Final = JaxReferencePoints
        Positions: Final = JaxPositions
        Headings: Final = JaxHeadings
        ObstacleStates: Final = JaxObstacleStates
        SampledObstacleStates: Final = JaxSampledObstacleStates
        Distance: Final = JaxDistance
        CostFunction: Final = JaxCostFunction
        PathParameterExtractor: Final = JaxPathParameterExtractor
        PathVelocityExtractor: Final = JaxPathVelocityExtractor
        PositionExtractor: Final = JaxPositionExtractor
        ContouringCost: Final = JaxContouringCost
        ObstacleStateProvider: Final = JaxObstacleStateProvider

        class simple:
            State: Final = JaxSimpleState
            StateSequence: Final = JaxSimpleStateSequence
            StateBatch: Final = JaxSimpleStateBatch
            ControlInputSequence: Final = JaxSimpleControlInputSequence
            ControlInputBatch: Final = JaxSimpleControlInputBatch
            Costs: Final = JaxSimpleCosts

        class integrator:
            ObstacleStateSequences: Final = JaxIntegratorObstacleStateSequences

        class bicycle:
            State: Final = JaxBicycleState
            StateSequence: Final = JaxBicycleStateSequence
            StateBatch: Final = JaxBicycleStateBatch
            Positions: Final = JaxBicyclePositions
            ControlInputSequence: Final = JaxBicycleControlInputSequence
            ControlInputBatch: Final = JaxBicycleControlInputBatch
            ObstacleStateSequences: Final = JaxBicycleObstacleStateSequences

        class augmented:
            State: Final = JaxAugmentedState
            StateSequence: Final = JaxAugmentedStateSequence
            StateBatch: Final = JaxAugmentedStateBatch
            ControlInputSequence: Final = JaxAugmentedControlInputSequence
            ControlInputBatch: Final = JaxAugmentedControlInputBatch



from typing import Final

from trajax.collectors.registry import CollectorRegistry
from trajax.collectors.obstacles import (
    ObstacleStateCollector,
    ObstacleForecastCollector,
)
from trajax.collectors.risk import RiskCollector
from trajax.collectors.state import (
    StateCollector,
    ControlCollector,
    TrajectoryCollector,
)


class collectors:
    states: Final = StateCollector
    controls: Final = ControlCollector
    trajectories: Final = TrajectoryCollector
    risk: Final = RiskCollector
    obstacle_states: Final = ObstacleStateCollector
    obstacle_forecasts: Final = ObstacleForecastCollector
    registry: Final = CollectorRegistry.of



from typing import Any
from dataclasses import dataclass
from functools import cached_property

from trajax.types import (
    Distance,
    DistanceExtractor,
    StateSequence,
    ObstacleStateSequences,
    SimulationData,
    Metric,
)
from trajax.collectors import access

from numtypes import Array, BoolArray, Dims, shape_of

import numpy as np


@dataclass(kw_only=True, frozen=True)
class CollisionMetricResult[T: int = int, V: int = int]:
    distances: Array[Dims[T, V]]
    distance_threshold: float

    @cached_property
    def min_distances(self) -> Array[Dims[V]]:
        return self.distances.min(axis=0)

    @cached_property
    def collisions(self) -> BoolArray[Dims[T, V]]:
        return self.distances <= self.distance_threshold

    @cached_property
    def collision_detected(self) -> bool:
        return bool(self.collisions.any())


@dataclass(kw_only=True, frozen=True)
class CollisionMetric[StateBatchT, SampledObstacleStatesT](
    Metric[CollisionMetricResult[Any]]
):
    distance: DistanceExtractor[StateBatchT, SampledObstacleStatesT, Distance]
    distance_threshold: float

    @staticmethod
    def create[S, SOS](
        *,
        distance_threshold: float,
        distance: DistanceExtractor[S, SOS, Distance],
    ) -> "CollisionMetric":
        return CollisionMetric(distance=distance, distance_threshold=distance_threshold)

    def compute[T: int = int](self, data: SimulationData) -> CollisionMetricResult[T]:
        states = data(
            access.states.assume(StateSequence[T, Any, StateBatchT]).require()
        )
        obstacle_states = data(
            access.obstacle_states.assume(
                ObstacleStateSequences[T, Any, Any, SampledObstacleStatesT]
            ).require()
        )

        measured_distances = self.distance(
            states=states.batched(), obstacle_states=obstacle_states.single()
        )

        distances = np.asarray(measured_distances).reshape(states.horizon, -1)

        assert shape_of(
            distances, matches=(states.horizon, measured_distances.vehicle_parts)
        )

        return CollisionMetricResult(
            distances=distances, distance_threshold=self.distance_threshold
        )

    @property
    def name(self) -> str:
        return "collision"



from typing import Final, cast
from dataclasses import dataclass

from trajax.types import (
    ObstacleIdAssignment,
    ObstacleStatesForTimeStep,
    ObstacleStatesHistory,
    NumPyObstaclePositionsForTimeStep,
    NumPyObstaclePositions,
    NumPyObstaclePositionExtractor,
)
from trajax.obstacles.history import NumPyObstacleIds

from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist
from numtypes import Array, IndexArray, BoolArray, Dims

import numpy as np


type PositionsExtractor[StatesT, HistoryT] = NumPyObstaclePositionExtractor[
    StatesT, HistoryT, NumPyObstaclePositionsForTimeStep, NumPyObstaclePositions
]


@dataclass(kw_only=True)
class NumPyHungarianObstacleIdAssignment[
    StatesT: ObstacleStatesForTimeStep,
    HistoryT: ObstacleStatesHistory,
](ObstacleIdAssignment[StatesT, NumPyObstacleIds, HistoryT]):
    positions: Final[PositionsExtractor[StatesT, HistoryT]]
    cutoff: Final[float]
    next_id: int

    @staticmethod
    def create[S: ObstacleStatesForTimeStep, H: ObstacleStatesHistory](
        *,
        position_extractor: PositionsExtractor[S, H],
        cutoff: float,
        start_id: int = 0,
    ) -> "NumPyHungarianObstacleIdAssignment":
        return NumPyHungarianObstacleIdAssignment(
            positions=position_extractor, cutoff=cutoff, next_id=start_id
        )

    def __call__(
        self, states: StatesT, /, *, history: HistoryT, ids: NumPyObstacleIds
    ) -> NumPyObstacleIds:
        if states.count == 0:
            return NumPyObstacleIds.empty()

        if history.horizon == 0:
            return NumPyObstacleIds.create(ids=self._allocate_ids(states.count))

        current_positions = self.positions.of_states_for_time_step(states).array
        last_positions, valid_ids = self._valid_history_from(history, ids)

        if len(valid_ids) == 0:
            return NumPyObstacleIds.create(ids=self._allocate_ids(states.count))

        current_indices, history_indices, matched = self._matching_for(
            current_positions=current_positions, last_positions=last_positions
        )

        return NumPyObstacleIds.create(
            ids=self._assign_ids(
                current_obstacle_count=states.count,
                current_indices=current_indices,
                history_indices=history_indices,
                matched=matched,
                valid_ids=valid_ids,
            )
        )

    def _valid_history_from[D_p: int = int, K: int = int](
        self, history: HistoryT, ids: NumPyObstacleIds
    ) -> tuple[Array[Dims[D_p, K]], IndexArray[Dims[K]]]:
        id_count = ids.count
        positions = self.positions.of_states(history).array

        # NOTE: The history may be padded with states for more obstacles
        # than there are IDs.
        last_positions = positions[-1, :, :id_count]

        # NOTE: Checking just the first dimension for nan is sufficient
        valid = ~np.isnan(last_positions[0])

        return last_positions[:, valid], ids.array[valid]

    def _matching_for[D_p: int = int, K_c: int = int, K_h: int = int, M: int = int](
        self,
        *,
        current_positions: Array[Dims[D_p, K_c]],
        last_positions: Array[Dims[D_p, K_h]],
    ) -> tuple[IndexArray[Dims[M]], IndexArray[Dims[M]], BoolArray[Dims[M]]]:
        distances = cdist(current_positions.T, last_positions.T)

        # NOTE: We add a large value to distances beyond the cutoff
        # to prevent them from being matched.
        gated_distances = distances.copy()
        gated_distances[gated_distances > self.cutoff] = self.cutoff + 1e9

        current_indices, history_indices = linear_sum_assignment(gated_distances)
        matched = distances[current_indices, history_indices] <= self.cutoff

        return current_indices, history_indices, matched

    def _assign_ids[K_c: int = int, M: int = int](
        self,
        *,
        current_obstacle_count: K_c,
        current_indices: IndexArray[Dims[M]],
        history_indices: IndexArray[Dims[M]],
        matched: BoolArray[Dims[M]],
        valid_ids: IndexArray,
    ) -> IndexArray[Dims[K_c]]:
        result = np.full(current_obstacle_count, -1, dtype=np.int64)
        result[current_indices[matched]] = valid_ids[history_indices[matched]]

        unmatched = result == -1
        result[unmatched] = self._allocate_ids(np.sum(unmatched))

        return cast(IndexArray[Dims[K_c]], result)

    def _allocate_ids[K: int = int](self, count: K) -> IndexArray[Dims[K]]:
        new_ids = np.arange(self.next_id, self.next_id + count)
        self.next_id += count

        return cast(IndexArray[Dims[K]], new_ids)



from typing import overload
from dataclasses import dataclass

from trajax.types import (
    DistanceExtractor,
    NumPyHeadings,
    NumPyPositions,
    NumPyPositionExtractor,
    NumPyHeadingExtractor,
)
from trajax.obstacles import NumPySampledObstacleStates
from trajax.costs.collision import NumPyDistance
from trajax.costs.distance.circles.common import Circles
from trajax.costs.distance.basic import replace_missing

from numtypes import Array, Dims, D, shape_of

import numpy as np


type OriginsArray[N: int = int] = Array[Dims[N, D[2]]]
type RadiiArray[N: int = int] = Array[Dims[N]]


@dataclass(frozen=True)
class NumPyCircleDistanceExtractor[StateT, V: int, C: int](
    DistanceExtractor[StateT, NumPySampledObstacleStates, NumPyDistance]
):
    """
    Computes the distances between parts of the ego robot and obstacles. Both the ego
    and the obstacles are represented as collections of circles.
    """

    ego: Circles[V]
    obstacle: Circles[C]
    positions_from: NumPyPositionExtractor[StateT]
    headings_from: NumPyHeadingExtractor[StateT]

    @staticmethod
    def create[S, V_: int, C_: int](
        *,
        ego: Circles[V_],
        obstacle: Circles[C_],
        position_extractor: NumPyPositionExtractor[S],
        heading_extractor: NumPyHeadingExtractor[S],
    ) -> "NumPyCircleDistanceExtractor[S, V_, C_]":
        return NumPyCircleDistanceExtractor(
            ego=ego,
            obstacle=obstacle,
            positions_from=position_extractor,
            headings_from=heading_extractor,
        )

    def __call__[T: int, N: int, M: int = int, K: int = int](
        self,
        *,
        states: StateT,
        obstacle_states: NumPySampledObstacleStates[T, K, N],
    ) -> NumPyDistance[T, V, M, N]:
        return NumPyDistance(
            compute_circle_distances(
                ego_positions=self.positions_from(states),
                ego_headings=self.headings_from(states),
                ego=self.ego,
                obstacle_states=replace_missing(obstacle_states),
                obstacle=self.obstacle,
            )
        )


def compute_circle_distances[T: int, M: int, V: int, C: int, K: int, N: int](
    *,
    ego_positions: NumPyPositions[T, M],
    ego_headings: NumPyHeadings[T, M],
    ego: Circles[V],
    obstacle_states: NumPySampledObstacleStates[T, K, N],
    obstacle: Circles[C],
) -> Array[Dims[T, V, M, N]]:
    ego_global_x, ego_global_y = to_global_positions(
        x=ego_positions.x(),
        y=ego_positions.y(),
        heading=ego_headings.heading,
        local_origins=ego.origins,
    )

    obstacle_global_x, obstacle_global_y = to_global_positions(
        x=obstacle_states.x(),
        y=obstacle_states.y(),
        heading=obstacle_states.heading(),
        local_origins=obstacle.origins,
    )

    pairwise_distances = pairwise_min_distances(
        ego_x=ego_global_x,
        ego_y=ego_global_y,
        ego_radii=ego.radii,
        obstacle_x=obstacle_global_x,
        obstacle_y=obstacle_global_y,
        obstacle_radii=obstacle.radii,
    )

    return min_distance_per_ego_part(pairwise_distances)


@overload
def to_global_positions[T: int, M: int, V: int](
    *,
    x: Array[Dims[T, M]],
    y: Array[Dims[T, M]],
    heading: Array[Dims[T, M]],
    local_origins: OriginsArray[V],
) -> tuple[Array[Dims[V, T, M]], Array[Dims[V, T, M]]]: ...


@overload
def to_global_positions[T: int, K: int, N: int, C: int](
    *,
    x: Array[Dims[T, K, N]],
    y: Array[Dims[T, K, N]],
    heading: Array[Dims[T, K, N]],
    local_origins: OriginsArray[C],
) -> tuple[Array[Dims[C, T, K, N]], Array[Dims[C, T, K, N]]]: ...


def to_global_positions[T: int, M: int, K: int, N: int, V: int, C: int](
    *,
    x: Array[Dims[T, M]] | Array[Dims[T, K, N]],
    y: Array[Dims[T, M]] | Array[Dims[T, K, N]],
    heading: Array[Dims[T, M]] | Array[Dims[T, K, N]],
    local_origins: OriginsArray[V] | OriginsArray[C],
) -> (
    tuple[Array[Dims[V, T, M]], Array[Dims[V, T, M]]]
    | tuple[Array[Dims[C, T, K, N]], Array[Dims[C, T, K, N]]]
):
    local_xy = local_origins.reshape((-1, 2) + (1,) * x.ndim)
    local_x, local_y = local_xy[:, 0], local_xy[:, 1]

    cos_h, sin_h = np.cos(heading), np.sin(heading)

    return (
        x + local_x * cos_h - local_y * sin_h,
        y + local_x * sin_h + local_y * cos_h,
    )


def pairwise_min_distances[T: int, M: int, V: int, C: int, K: int, N: int](
    ego_x: Array[Dims[V, T, M]],
    ego_y: Array[Dims[V, T, M]],
    ego_radii: RadiiArray[V],
    obstacle_x: Array[Dims[C, T, K, N]],
    obstacle_y: Array[Dims[C, T, K, N]],
    obstacle_radii: RadiiArray[C],
) -> Array[Dims[V, C, T, M, K, N]]:
    dx = (
        ego_x[:, np.newaxis, :, :, np.newaxis, np.newaxis]
        - obstacle_x[np.newaxis, :, :, np.newaxis, :, :]
    )
    dy = (
        ego_y[:, np.newaxis, :, :, np.newaxis, np.newaxis]
        - obstacle_y[np.newaxis, :, :, np.newaxis, :, :]
    )

    center_distances = np.sqrt(dx**2 + dy**2)
    radii_sum = (
        ego_radii[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
        + obstacle_radii[np.newaxis, :, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
    )

    return center_distances - radii_sum


def min_distance_per_ego_part[T: int, M: int, V: int, C: int, K: int, N: int](
    pairwise_distances: Array[Dims[V, C, T, M, K, N]],
) -> Array[Dims[T, V, M, N]]:
    V, C, T, M, K, N = pairwise_distances.shape

    if C == 0 or K == 0:
        distances = np.full((T, V, M, N), np.inf)

        assert shape_of(distances, matches=(T, V, M, N))

        return distances

    min_over_obstacles = np.min(pairwise_distances, axis=(1, 4))
    return np.transpose(min_over_obstacles, (1, 0, 2, 3))



import numpy as np

from trajax.obstacles import NumPySampledObstacleStates


def replace_missing[T: int, K: int, N: int](
    obstacle_states: NumPySampledObstacleStates[T, K, N],
) -> NumPySampledObstacleStates[T, K, N]:
    return NumPySampledObstacleStates.create(
        x=np.nan_to_num(obstacle_states.x(), nan=np.inf, posinf=np.inf, neginf=np.inf),
        y=np.nan_to_num(obstacle_states.y(), nan=np.inf, posinf=np.inf, neginf=np.inf),
        heading=np.nan_to_num(
            obstacle_states.heading(), nan=0.0, posinf=0.0, neginf=0.0
        ),
    )



from dataclasses import dataclass

from trajax.types import NumPyControlInputSequence

from numtypes import Array, Dim1, shape_of
from scipy.signal import savgol_coeffs
from scipy.ndimage import convolve1d


@dataclass(frozen=True)
class NumPySavGolFilter:
    coefficients: Array[Dim1]

    @staticmethod
    def create(*, window_length: int, polynomial_order: int) -> "NumPySavGolFilter":
        assert window_length % 2 == 1, f"Window length must be odd, got {window_length}"
        assert 0 <= polynomial_order < window_length, (
            f"Polynomial order must be non-negative and less than window length, got {polynomial_order}"
        )

        coefficients = savgol_coeffs(window_length, polynomial_order)

        assert shape_of(coefficients, matches=(window_length,))

        return NumPySavGolFilter(coefficients=coefficients)

    def __call__[InputSequenceT: NumPyControlInputSequence](
        self, *, optimal_input: InputSequenceT
    ) -> InputSequenceT:
        filtered = convolve1d(
            optimal_input.array, weights=self.coefficients, axis=0, mode="nearest"
        )

        assert shape_of(filtered, matches=optimal_input.array.shape)

        return optimal_input.similar(array=filtered)



from typing import Final


from trajax.obstacles.common import PredictingObstacleStateProvider
from trajax.obstacles.static import (
    NumPyStaticObstacleSimulator,
    JaxStaticObstacleSimulator,
)
from trajax.obstacles.dynamic import (
    NumPyDynamicObstacleSimulator,
    JaxDynamicObstacleSimulator,
)
from trajax.obstacles.sampler import (
    NumPyGaussianObstacleStateSampler,
    JaxGaussianObstacleStateSampler,
)
from trajax.obstacles.assignment import (
    NumPyHungarianObstacleIdAssignment,
    JaxHungarianObstacleIdAssignment,
)


class obstacles:
    class numpy:
        empty: Final = NumPyStaticObstacleSimulator.empty
        static: Final = NumPyStaticObstacleSimulator.create
        dynamic: Final = NumPyDynamicObstacleSimulator.create

        class provider:
            predicting: Final = PredictingObstacleStateProvider.create

        class sampler:
            gaussian: Final = NumPyGaussianObstacleStateSampler.create

        class id_assignment:
            hungarian: Final = NumPyHungarianObstacleIdAssignment.create

    class jax:
        empty: Final = JaxStaticObstacleSimulator.empty
        static: Final = JaxStaticObstacleSimulator.create
        dynamic: Final = JaxDynamicObstacleSimulator.create

        class provider:
            predicting: Final = PredictingObstacleStateProvider.create

        class sampler:
            gaussian: Final = JaxGaussianObstacleStateSampler.create

        class id_assignment:
            hungarian: Final = JaxHungarianObstacleIdAssignment.create



from typing import Protocol, Final

from trajax.types.array import DataType

from numtypes import Array, Dims, D

BICYCLE_D_X: Final = 4
BICYCLE_D_V: Final = 2
BICYCLE_D_U: Final = 2
BICYCLE_D_O: Final = 3

type BicycleD_x = D[4]
type BicycleD_v = D[2]
type BicycleD_u = D[2]
type BicycleD_o = D[3]


class BicycleState(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[BicycleD_x]]:
        """Returns the state as a NumPy array."""
        ...

    @property
    def x(self) -> float:
        """X position of the agent."""
        ...

    @property
    def y(self) -> float:
        """Y position of the agent."""
        ...

    @property
    def heading(self) -> float:
        """Orientation of the agent."""
        ...

    @property
    def speed(self) -> float:
        """Velocity of the agent."""
        ...

    @property
    def dimension(self) -> BicycleD_x:
        """State dimension."""
        ...


class BicycleStateSequence(Protocol):
    def step(self, index: int) -> BicycleState:
        """Returns the state at the given time step index."""
        ...


class BicycleStateBatch[T: int, M: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, BicycleD_x, M]]:
        """Returns the states as a NumPy array."""
        ...

    def heading(self) -> Array[Dims[T, M]]:
        """Returns the headings (orientations) of the states in the batch."""
        ...

    def speed(self) -> Array[Dims[T, M]]:
        """Returns the speeds of the states in the batch."""
        ...

    def rollout(self, index: int) -> BicycleStateSequence:
        """Returns a single rollout from the batch as a state sequence."""
        ...

    @property
    def positions(self) -> "BicyclePositions[T, M] ":
        """Returns the positions of the states in the batch."""
        ...


class BicyclePositions[T: int, M: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, BicycleD_u, M]]:
        """Returns the positions as a NumPy array."""
        ...

    def x(self) -> Array[Dims[T, M]]:
        """Returns the x positions."""
        ...

    def y(self) -> Array[Dims[T, M]]:
        """Returns the y positions."""
        ...


class BicycleControlInputSequence[T: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, BicycleD_u]]:
        """Returns the control input sequence as a NumPy array."""
        ...

    @property
    def horizon(self) -> T:
        """Time horizon of the control input sequence."""
        ...

    @property
    def dimension(self) -> BicycleD_u:
        """Control input dimension."""
        ...


class BicycleControlInputBatch[T: int, M: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, BicycleD_u, M]]:
        """Returns the control inputs as a NumPy array."""
        ...

    @property
    def rollout_count(self) -> M:
        """Number of rollouts in the batch."""
        ...

    @property
    def horizon(self) -> T:
        """Time horizon of the control inputs."""
        ...



from typing import Protocol, Self, Any, overload

from trajax.types.mppi.common import (
    State,
    StateSequence,
    StateBatch,
    ControlInputSequence,
    ControlInputBatch,
    DynamicalModel,
    Costs,
    CostFunction,
    Sampler,
    UpdateFunction,
    PaddingFunction,
    FilterFunction,
)

from numtypes import Array, Dims


class NumPyState[D_x: int](State[D_x], Protocol):
    @property
    def array(self) -> Array[Dims[D_x]]:
        """Returns the underlying NumPy array representing the state."""
        ...


class NumPyStateSequence[T: int, D_x: int, StateBatchT = Any](
    StateSequence[T, D_x, StateBatchT], Protocol
):
    @property
    def array(self) -> Array[Dims[T, D_x]]:
        """Returns the underlying NumPy array representing the state sequence."""
        ...


class NumPyStateBatch[T: int, D_x: int, M: int](StateBatch[T, D_x, M], Protocol):
    @property
    def array(self) -> Array[Dims[T, D_x, M]]:
        """Returns the underlying NumPy array representing the state batch."""
        ...


class NumPyControlInputSequence[T: int, D_u: int](
    ControlInputSequence[T, D_u], Protocol
):
    @overload
    def similar(self, *, array: Array[Dims[T, D_u]]) -> Self:
        """Creates a new control input sequence similar to this one but with the given
        array as its data.
        """
        ...

    @overload
    def similar[L: int](
        self, *, array: Array[Dims[L, D_u]], length: L
    ) -> "NumPyControlInputSequence[L, D_u]":
        """Creates a new control input sequence similar to this one but with the given
        array as its data. The length of the new sequence may differ from the original.
        """
        ...

    @property
    def array(self) -> Array[Dims[T, D_u]]:
        """Returns the underlying NumPy array representing the control input sequence."""
        ...


class NumPyControlInputBatch[T: int, D_u: int, M: int](
    ControlInputBatch[T, D_u, M], Protocol
):
    @property
    def array(self) -> Array[Dims[T, D_u, M]]:
        """Returns the underlying NumPy array representing the control input batch."""
        ...


class NumPyCosts[T: int, M: int](Costs[T, M], Protocol):
    def similar(self, *, array: Array[Dims[T, M]]) -> Self:
        """Creates new costs similar to this one but with the given array as its data."""
        ...

    @property
    def array(self) -> Array[Dims[T, M]]:
        """Returns the underlying NumPy array representing the costs."""
        ...


class NumPyDynamicalModel[
    StateT,
    StateSequenceT,
    StateBatchT,
    InputSequenceT,
    InputBatchT,
](
    DynamicalModel[StateT, StateSequenceT, StateBatchT, InputSequenceT, InputBatchT],
    Protocol,
): ...


class NumPySampler[InputSequenceT, InputBatchT](
    Sampler[InputSequenceT, InputBatchT], Protocol
): ...


class NumPyCostFunction[InputBatchT, StateBatchT, CostsT](
    CostFunction[InputBatchT, StateBatchT, CostsT], Protocol
): ...


class NumPyUpdateFunction[InputSequenceT](UpdateFunction[InputSequenceT], Protocol): ...


class NumPyPaddingFunction[NominalT, PaddingT](
    PaddingFunction[NominalT, PaddingT], Protocol
): ...


class NumPyFilterFunction[InputSequenceT](FilterFunction[InputSequenceT], Protocol): ...



from dataclasses import dataclass

from numtypes import Array, Dims, D

type OriginsArray[V: int] = Array[Dims[V, D[2]]]
type RadiiArray[V: int] = Array[Dims[V]]


@dataclass(frozen=True)
class Circles[V: int]:
    """Describes circles approximating parts of an object for distance computation."""

    origins: OriginsArray[V]
    """The local (x, y) offset of the circle center from the object center."""

    radii: RadiiArray[V]
    """The radius of the circle."""



from typing import Protocol, Self


class ObstacleSimulator[ObstacleStatesForTimeStepT](Protocol):
    def step(self) -> ObstacleStatesForTimeStepT:
        """Advances the obstacle simulation by one time step and returns the new states."""
        ...

    def with_time_step_size(self, time_step_size: float) -> Self:
        """Returns a new simulator with the specified time step size."""
        ...

    @property
    def obstacle_count(self) -> int:
        """Returns the number of obstacles being simulated."""
        ...



from typing import Sequence, cast
from dataclasses import dataclass
from functools import cached_property

from trajax.types import DataType, D_o, D_O, SampledObstacleStates, ObstacleStates

from numtypes import Array, Dims, D, shape_of

import numpy as np


type ObstacleCovarianceArray[T: int = int, K: int = int] = Array[Dims[T, D_o, D_o, K]]


@dataclass(kw_only=True, frozen=True)
class NumPySampledObstacleStates[T: int, K: int, N: int](
    SampledObstacleStates[T, K, N]
):
    _x: Array[Dims[T, K, N]]
    _y: Array[Dims[T, K, N]]
    _heading: Array[Dims[T, K, N]]

    @staticmethod
    def create[T_: int, K_: int, N_: int](
        *,
        x: Array[Dims[T_, K_, N_]],
        y: Array[Dims[T_, K_, N_]],
        heading: Array[Dims[T_, K_, N_]],
    ) -> "NumPySampledObstacleStates[T_, K_, N_]":
        return NumPySampledObstacleStates(_x=x, _y=y, _heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_o, K, N]]:
        return self.array

    def x(self) -> Array[Dims[T, K, N]]:
        return self._x

    def y(self) -> Array[Dims[T, K, N]]:
        return self._y

    def heading(self) -> Array[Dims[T, K, N]]:
        return self._heading

    def at(self, *, time_step: int, sample: int) -> "NumPyObstacleStatesForTimeStep[K]":
        return NumPyObstacleStatesForTimeStep.create(
            x=self._x[time_step, :, sample],
            y=self._y[time_step, :, sample],
            heading=self._heading[time_step, :, sample],
        )

    @property
    def horizon(self) -> T:
        return self._x.shape[0]

    @property
    def dimension(self) -> D_o:
        return D_O

    @property
    def count(self) -> K:
        return self._x.shape[1]

    @property
    def sample_count(self) -> N:
        return self._x.shape[2]

    @property
    def array(self) -> Array[Dims[T, D_o, K, N]]:
        return self._array

    @cached_property
    def _array(self) -> Array[Dims[T, D_o, K, N]]:
        return np.stack([self._x, self._y, self._heading], axis=1)


@dataclass(kw_only=True, frozen=True)
class NumPyObstacle2dPositions[T: int, K: int]:
    _x: Array[Dims[T, K]]
    _y: Array[Dims[T, K]]

    @staticmethod
    def create[T_: int, K_: int](
        *, x: Array[Dims[T_, K_]], y: Array[Dims[T_, K_]]
    ) -> "NumPyObstacle2dPositions[T_, K_]":
        return NumPyObstacle2dPositions(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D[2], K]]:
        return self.array

    @property
    def horizon(self) -> T:
        return self._x.shape[0]

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> K:
        return self._x.shape[1]

    @property
    def array(self) -> Array[Dims[T, D[2], K]]:
        return self._array

    @cached_property
    def _array(self) -> Array[Dims[T, D[2], K]]:
        return np.stack([self._x, self._y], axis=1)


@dataclass(kw_only=True, frozen=True)
class NumPyObstacleStates[T: int, K: int](
    ObstacleStates[T, K, NumPySampledObstacleStates[T, K, D[1]]]
):
    _x: Array[Dims[T, K]]
    _y: Array[Dims[T, K]]
    _heading: Array[Dims[T, K]]
    _covariance: ObstacleCovarianceArray[T, K] | None

    @staticmethod
    def empty[T_: int, K_: int = D[0]](
        *, horizon: T_, obstacle_count: K_ = 0
    ) -> "NumPyObstacleStates[T_, K_]":
        """Creates obstacle states for zero obstacles over the given time horizon."""
        empty = np.full((horizon, obstacle_count), fill_value=np.nan)

        assert shape_of(empty, matches=(horizon, obstacle_count))

        return NumPyObstacleStates.create(x=empty, y=empty, heading=empty)

    @staticmethod
    def wrap[T_: int, K_: int](
        array: Array[Dims[T_, D_o, K_]],
    ) -> "NumPyObstacleStates[T_, K_]":
        return NumPyObstacleStates.create(
            x=array[:, 0, :], y=array[:, 1, :], heading=array[:, 2, :]
        )

    @staticmethod
    def sampled[T_: int, K_: int, N_: int](  # type: ignore
        *,
        x: Array[Dims[T_, K_, N_]],
        y: Array[Dims[T_, K_, N_]],
        heading: Array[Dims[T_, K_, N_]],
    ) -> NumPySampledObstacleStates[T_, K_, N_]:
        return NumPySampledObstacleStates.create(x=x, y=y, heading=heading)

    @staticmethod
    def create[T_: int, K_: int](
        *,
        x: Array[Dims[T_, K_]],
        y: Array[Dims[T_, K_]],
        heading: Array[Dims[T_, K_]],
        covariance: ObstacleCovarianceArray[T_, K_] | None = None,
    ) -> "NumPyObstacleStates[T_, K_]":
        return NumPyObstacleStates(_x=x, _y=y, _heading=heading, _covariance=covariance)

    @staticmethod
    def of_states[T_: int, K_: int](
        obstacle_states: Sequence["NumPyObstacleStatesForTimeStep[K_]"],
        *,
        horizon: T_ | None = None,
    ) -> "NumPyObstacleStates[T_, K_]":
        assert len(obstacle_states) > 0, "Obstacle states sequence must not be empty."
        assert horizon is None or len(obstacle_states) == horizon, (
            f"Expected horizon {horizon}, but got {len(obstacle_states)} obstacle states."
        )

        T = len(obstacle_states)
        K = max(states.count for states in obstacle_states)

        x = np.full((T, K), np.nan)
        y = np.full((T, K), np.nan)
        heading = np.full((T, K), np.nan)

        for t, states in enumerate(obstacle_states):
            n = states.count
            x[t, :n] = states.x()
            y[t, :n] = states.y()
            heading[t, :n] = states.heading()

        return cast(
            NumPyObstacleStates[T_, K_],
            NumPyObstacleStates.create(x=x, y=y, heading=heading),
        )

    @staticmethod
    def for_time_step[K_: int](
        *, x: Array[Dims[K_]], y: Array[Dims[K_]], heading: Array[Dims[K_]]
    ) -> "NumPyObstacleStatesForTimeStep[K_]":
        return NumPyObstacleStatesForTimeStep.create(x=x, y=y, heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_o, K]]:
        return self.array

    def x(self) -> Array[Dims[T, K]]:
        return self._x

    def y(self) -> Array[Dims[T, K]]:
        return self._y

    def heading(self) -> Array[Dims[T, K]]:
        return self._heading

    def positions(self) -> NumPyObstacle2dPositions[T, K]:
        return NumPyObstacle2dPositions.create(x=self._x, y=self._y)

    def covariance(self) -> ObstacleCovarianceArray[T, K] | None:
        return self._covariance

    def single(self) -> NumPySampledObstacleStates[T, K, D[1]]:
        return NumPySampledObstacleStates.create(
            x=self._x[..., np.newaxis],
            y=self._y[..., np.newaxis],
            heading=self._heading[..., np.newaxis],
        )

    def last(self) -> "NumPyObstacleStatesForTimeStep[K]":
        return self.at(time_step=self.horizon - 1)

    def at(self, *, time_step: int) -> "NumPyObstacleStatesForTimeStep[K]":
        return NumPyObstacleStatesForTimeStep.create(
            x=self._x[time_step],
            y=self._y[time_step],
            heading=self._heading[time_step],
        )

    @property
    def horizon(self) -> T:
        return self._x.shape[0]

    @property
    def dimension(self) -> D_o:
        return D_O

    @property
    def count(self) -> K:
        return self._x.shape[1]

    @property
    def array(self) -> Array[Dims[T, D_o, K]]:
        return self._array

    @cached_property
    def _array(self) -> Array[Dims[T, D_o, K]]:
        return np.stack([self._x, self._y, self._heading], axis=1)


@dataclass(kw_only=True, frozen=True)
class NumPyObstacle2dPositionsForTimeStep[K: int]:
    _x: Array[Dims[K]]
    _y: Array[Dims[K]]

    @staticmethod
    def create[K_: int](
        *, x: Array[Dims[K_]], y: Array[Dims[K_]]
    ) -> "NumPyObstacle2dPositionsForTimeStep[K_]":
        return NumPyObstacle2dPositionsForTimeStep(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D[2], K]]:
        return self.array

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> K:
        return self._x.shape[0]

    @property
    def array(self) -> Array[Dims[D[2], K]]:
        return self._array

    @cached_property
    def _array(self) -> Array[Dims[D[2], K]]:
        return np.stack([self._x, self._y], axis=0)


@dataclass(kw_only=True, frozen=True)
class NumPyObstacleStatesForTimeStep[K: int]:
    _x: Array[Dims[K]]
    _y: Array[Dims[K]]
    _heading: Array[Dims[K]]

    @staticmethod
    def create[K_: int](
        *,
        x: Array[Dims[K_]],
        y: Array[Dims[K_]],
        heading: Array[Dims[K_]],
    ) -> "NumPyObstacleStatesForTimeStep[K_]":
        return NumPyObstacleStatesForTimeStep(_x=x, _y=y, _heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D_o, K]]:
        return self.array

    def x(self) -> Array[Dims[K]]:
        return self._x

    def y(self) -> Array[Dims[K]]:
        return self._y

    def heading(self) -> Array[Dims[K]]:
        return self._heading

    def positions(self) -> NumPyObstacle2dPositionsForTimeStep[K]:
        return NumPyObstacle2dPositionsForTimeStep.create(x=self._x, y=self._y)

    def replicate[T: int](self, *, horizon: T) -> NumPyObstacleStates[T, K]:
        return NumPyObstacleStates.create(
            x=np.tile(self._x[np.newaxis, :], (horizon, 1)),
            y=np.tile(self._y[np.newaxis, :], (horizon, 1)),
            heading=np.tile(self._heading[np.newaxis, :], (horizon, 1)),
        )

    @property
    def dimension(self) -> D_o:
        return D_O

    @property
    def count(self) -> K:
        return self._x.shape[0]

    @property
    def array(self) -> Array[Dims[D_o, K]]:
        return self._array

    @cached_property
    def _array(self) -> Array[Dims[D_o, K]]:
        return np.stack([self._x, self._y, self._heading], axis=0)



from .basic import (
    NumPySimpleState as NumPySimpleState,
    NumPySimpleStateSequence as NumPySimpleStateSequence,
    NumPySimpleStateBatch as NumPySimpleStateBatch,
    NumPySimpleControlInputSequence as NumPySimpleControlInputSequence,
    NumPySimpleControlInputBatch as NumPySimpleControlInputBatch,
    NumPySimpleCosts as NumPySimpleCosts,
)
from .accelerated import (
    JaxSimpleState as JaxSimpleState,
    JaxSimpleStateSequence as JaxSimpleStateSequence,
    JaxSimpleStateBatch as JaxSimpleStateBatch,
    JaxSimpleControlInputSequence as JaxSimpleControlInputSequence,
    JaxSimpleControlInputBatch as JaxSimpleControlInputBatch,
    JaxSimpleCosts as JaxSimpleCosts,
)



from typing import Any
from dataclasses import dataclass

from trajax.types import (
    jaxtyped,
    ObstacleStateSequences,
    JaxInitialPositionCovariance,
    JaxInitialVelocityCovariance,
)

from numtypes import Array, Dims
from jaxtyping import Array as JaxArray, Float, Scalar

import jax.numpy as jnp


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxConstantVarianceProvider:
    position_variance: Scalar
    velocity_variance: Scalar

    @staticmethod
    def create(
        *,
        position_variance: float | Scalar,
        velocity_variance: float | Scalar,
    ) -> "JaxConstantVarianceProvider":
        return JaxConstantVarianceProvider(
            position_variance=jnp.asarray(position_variance),
            velocity_variance=jnp.asarray(velocity_variance),
        )

    def position[K: int](
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> JaxInitialPositionCovariance[K]:
        return jnp.tile(
            (jnp.eye(2) * self.position_variance)[..., jnp.newaxis],
            (1, 1, states.count),
        )

    def velocity[K: int](
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> JaxInitialVelocityCovariance[K]:
        return jnp.tile(
            (jnp.eye(2) * self.velocity_variance)[..., jnp.newaxis],
            (1, 1, states.count),
        )


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxConstantCovarianceProvider[K: int]:
    position_covariance: Float[JaxArray, "2 2 K"]
    velocity_covariance: Float[JaxArray, "2 2 K"]

    @staticmethod
    def create(
        *,
        position_covariance: Array[Dims[K]] | Float[JaxArray, "2 2 K"],
        velocity_covariance: Array[Dims[K]] | Float[JaxArray, "2 2 K"],
    ) -> "JaxConstantCovarianceProvider[K]":
        return JaxConstantCovarianceProvider(
            position_covariance=jnp.asarray(position_covariance),
            velocity_covariance=jnp.asarray(velocity_covariance),
        )

    def position(
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> JaxInitialPositionCovariance[K]:
        assert states.count == self.position_covariance.shape[2]
        return self.position_covariance

    def velocity(
        self, states: ObstacleStateSequences[int, K, Any]
    ) -> JaxInitialVelocityCovariance[K]:
        assert states.count == self.velocity_covariance.shape[2]
        return self.velocity_covariance



from typing import cast, Self
from dataclasses import dataclass
from functools import cached_property

from trajax.types import DataType, jaxtyped
from trajax.obstacles.accelerated import JaxObstacleStatesForTimeStep, JaxObstacleStates
from trajax.obstacles.history.basic import (
    NumPyObstacleIds,
    NumPyObstacleStatesRunningHistory,
)

from numtypes import IndexArray, Dims
from jaxtyping import Array as JaxArray, Int, Num

import numpy as np
import jax.numpy as jnp


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class JaxObstacleIds[K: int]:
    _ids: Int[JaxArray, "K"]

    @staticmethod
    def create[K_: int](
        *,
        ids: Num[JaxArray, "K"],
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacleIds[K_]":
        return JaxObstacleIds(_ids=ids.astype(jnp.int32))

    def __array__(self, dtype: DataType | None = None) -> IndexArray[Dims[K]]:
        return self._numpy_array

    def numpy(self) -> NumPyObstacleIds[K]:
        return NumPyObstacleIds.create(ids=self._numpy_array)

    @property
    def count(self) -> K:
        return cast(K, self._ids.shape[0])

    @property
    def array(self) -> Int[JaxArray, "K"]:
        return self._ids

    @property
    def _numpy_array(self) -> IndexArray[Dims[K]]:
        return np.asarray(self._ids, dtype=np.intp)


@dataclass(kw_only=True, frozen=True)
class JaxObstacleStatesRunningHistory[H: int, K: int]:
    # NOTE: Internally uses NumPy implementation, since it's faster in most cases.
    history: NumPyObstacleStatesRunningHistory[H, K]

    @staticmethod
    def empty[H_: int = int, K_: int = int](
        *, horizon: H_ | None = None, obstacle_count: K_ | None = None
    ) -> "JaxObstacleStatesRunningHistory[H_, K_]":
        return JaxObstacleStatesRunningHistory(
            history=NumPyObstacleStatesRunningHistory.empty(
                horizon=horizon, obstacle_count=obstacle_count
            )
        )

    @staticmethod
    def single[H_: int = int, K_: int = int](
        observation: JaxObstacleStatesForTimeStep,
        *,
        horizon: H_ | None = None,
        obstacle_count: K_ | None = None,
    ) -> "JaxObstacleStatesRunningHistory[H_, K_]":
        return JaxObstacleStatesRunningHistory(
            history=NumPyObstacleStatesRunningHistory.single(
                observation=observation.numpy(),
                horizon=horizon,
                obstacle_count=obstacle_count,
            )
        )

    def last(self) -> JaxObstacleStatesForTimeStep[K]:
        return self._get.last()

    def get(self) -> JaxObstacleStates[int, K]:
        return self._get

    def ids(self) -> JaxObstacleIds[K]:
        return self._ids

    def append[N: int](
        self,
        observation: JaxObstacleStatesForTimeStep[N],
        *,
        ids: "JaxObstacleIds[N] | None" = None,
    ) -> Self:
        return self.__class__(
            history=self.history.append(
                observation=observation.numpy(),
                ids=ids.numpy() if ids is not None else None,
            )
        )

    @property
    def horizon(self) -> int:
        return self.history.horizon

    @property
    def count(self) -> K:
        return self.history.count

    @cached_property
    def _get(self) -> JaxObstacleStates[int, K]:
        return JaxObstacleStates.wrap(jnp.asarray(self.history.get().array))

    @cached_property
    def _ids(self) -> JaxObstacleIds[K]:
        return JaxObstacleIds.create(
            ids=jnp.asarray(self.history.ids().array), obstacle_count=self.count
        )



from typing import Protocol
from dataclasses import dataclass

from trajax.types.array import DataType
from trajax.types.costs.collision.common import (
    ObstacleStateProvider,
    ObstacleStateSampler,
    DistanceExtractor,
    SampleCostFunction,
    Risk,
    RiskMetric,
)

from numtypes import Array, Dims


class NumPyObstacleStateProvider[ObstacleStatesT](
    ObstacleStateProvider[ObstacleStatesT], Protocol
): ...


class NumPyObstacleStateSampler[ObstacleStatesT, SampledObstacleStatesT](
    ObstacleStateSampler[ObstacleStatesT, SampledObstacleStatesT], Protocol
): ...


class NumPyDistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT](
    DistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT], Protocol
): ...


@dataclass(frozen=True)
class NumPyRisk[T: int, M: int](Risk[T, M]):
    _array: Array[Dims[T, M]]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        return self._array

    @property
    def horizon(self) -> T:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self._array.shape[1]

    @property
    def array(self) -> Array[Dims[T, M]]:
        return self._array


class NumPyRiskMetric[StateBatchT, ObstacleStatesT, SampledObstacleStatesT](
    RiskMetric, Protocol
):
    def compute[T: int, M: int](
        self,
        cost_function: SampleCostFunction[
            StateBatchT, SampledObstacleStatesT, Array[Dims[T, M, int]]
        ],
        *,
        states: StateBatchT,
        obstacle_states: ObstacleStatesT,
        sampler: NumPyObstacleStateSampler[ObstacleStatesT, SampledObstacleStatesT],
    ) -> NumPyRisk[T, M]:
        """Computes the risk metric based on the provided cost function and returns it as a NumPy array."""
        ...



from dataclasses import dataclass
from functools import cached_property

from trajax.types import (
    Trajectory,
    NumPyPathParameters,
    NumPyReferencePoints,
    NumPyPositions,
    NumPyLateralPositions,
    NumPyLongitudinalPositions,
    NumPyNormals,
)

from numtypes import Array, Dims, D, shape_of

import numpy as np


type Vector = Array[Dims[D[2]]]


@dataclass(kw_only=True, frozen=True)
class NumPyLineTrajectory(
    Trajectory[
        NumPyPathParameters,
        NumPyReferencePoints,
        NumPyPositions,
        NumPyLateralPositions,
        NumPyLongitudinalPositions,
    ]
):
    start: Vector
    direction: Vector

    heading: float

    _end: tuple[float, float]
    _path_length: float

    @staticmethod
    def create(
        *, start: tuple[float, float], end: tuple[float, float], path_length: float
    ) -> "NumPyLineTrajectory":
        """Generates a straight line trajectory from start to end."""
        return NumPyLineTrajectory(
            start=(start_array := np.array(start)),
            direction=(direction := np.array(end) - start_array),
            heading=np.arctan2(direction[1], direction[0]),
            _end=end,
            _path_length=path_length,
        )

    def query[T: int, M: int](
        self, parameters: NumPyPathParameters[T, M]
    ) -> NumPyReferencePoints[T, M]:
        T, M = parameters.horizon, parameters.rollout_count
        normalized = parameters.array / self.path_length

        x, y = (
            self.start[:, np.newaxis, np.newaxis]
            + self.direction[:, np.newaxis, np.newaxis] * normalized
        )
        heading = np.full_like(x, self.heading)

        assert shape_of(x, matches=(T, M), name="x")
        assert shape_of(y, matches=(T, M), name="y")
        assert shape_of(heading, matches=(T, M), name="heading")

        return NumPyReferencePoints.create(x=x, y=y, heading=heading)

    def lateral[T: int, M: int](
        self, positions: NumPyPositions[T, M]
    ) -> NumPyLateralPositions[T, M]:
        T, M = positions.horizon, positions.rollout_count
        relative = positions.array - self.start[:, np.newaxis]

        assert shape_of(relative, matches=(T, 2, M), name="relative")

        lateral = np.einsum("tpm,p->tm", relative, self.perpendicular)

        assert shape_of(lateral, matches=(T, M), name="lateral")

        return NumPyLateralPositions.create(lateral)

    def longitudinal[T: int, M: int](
        self, positions: NumPyPositions[T, M]
    ) -> NumPyLongitudinalPositions[T, M]:
        T, M = positions.horizon, positions.rollout_count
        relative = positions.array - self.start[:, np.newaxis]

        assert shape_of(relative, matches=(T, 2, M), name="relative")

        projection = np.einsum("tpm,p->tm", relative, self.tangent)
        longitudinal = projection * self.path_length / self.line_length

        assert shape_of(longitudinal, matches=(T, M), name="longitudinal")

        return NumPyLongitudinalPositions.create(longitudinal)

    def normal[T: int, M: int](
        self, parameters: NumPyPathParameters[T, M]
    ) -> NumPyNormals[T, M]:
        T, M = parameters.horizon, parameters.rollout_count

        x = np.full((T, M), self.perpendicular[0])
        y = np.full((T, M), self.perpendicular[1])

        assert shape_of(x, matches=(T, M), name="normal x")
        assert shape_of(y, matches=(T, M), name="normal y")

        return NumPyNormals.create(x=x, y=y)

    @property
    def end(self) -> tuple[float, float]:
        return self._end

    @property
    def path_length(self) -> float:
        return self._path_length

    @property
    def natural_length(self) -> float:
        return self.line_length

    @cached_property
    def perpendicular(self) -> Vector:
        tangent = self.tangent
        perpendicular = np.array([tangent[1], -tangent[0]])

        assert shape_of(perpendicular, matches=(2,), name="perpendicular")

        return perpendicular

    @cached_property
    def tangent(self) -> Vector:
        tangent = self.direction / self.line_length

        assert shape_of(tangent, matches=(2,), name="tangent")

        return tangent

    @cached_property
    def line_length(self) -> float:
        return float(np.linalg.norm(self.direction))



from typing import cast
from dataclasses import dataclass

from trajax.types import JaxObstacleStateSampler
from trajax.obstacles.accelerated import JaxSampledObstacleStates, JaxObstacleStates

from jaxtyping import PRNGKeyArray
from riskit import distribution

import jax
import jax.numpy as jnp


@dataclass
class JaxGaussianObstacleStateSampler(
    JaxObstacleStateSampler[JaxObstacleStates, JaxSampledObstacleStates]
):
    key: PRNGKeyArray

    @staticmethod
    def create(*, seed: int = 42) -> "JaxGaussianObstacleStateSampler":
        return JaxGaussianObstacleStateSampler(key=jax.random.key(seed))

    def __call__[T: int, K: int, N: int](
        self, states: JaxObstacleStates[T, K], *, count: N
    ) -> JaxSampledObstacleStates[T, K, N]:
        if states.count == 0:
            return cast(JaxSampledObstacleStates, states.single())

        if (covariance := states.covariance_array) is None:
            assert count == 1, (
                "It's pointless to take multiple samples, when covariance information is not available."
            )
            return cast(JaxSampledObstacleStates, states.single())

        T, D_O, _, K = covariance.shape

        mean = jnp.stack([states.x_array, states.y_array, states.heading_array], axis=1)
        flat_covariance = covariance.transpose(0, 3, 1, 2).reshape(-1, D_O, D_O)
        flat_mean = mean.transpose(0, 2, 1).reshape(-1, D_O)

        gaussian = distribution.jax.gaussian(
            mean=flat_mean, covariance=flat_covariance, key=self.key
        )
        samples = gaussian.sample(count=count)
        self.key = gaussian.key

        samples = samples.reshape(T, K, D_O, count).transpose(0, 2, 1, 3)

        return states.sampled(
            x=samples[:, 0, :, :],
            y=samples[:, 1, :, :],
            heading=samples[:, 2, :, :],  # type: ignore
            sample_count=count,
        )



from typing import Final

from trajax.samplers.gaussian import NumPyGaussianSampler, JaxGaussianSampler
from trajax.samplers.halton import NumPyHaltonSplineSampler, JaxHaltonSplineSampler


class sampler:
    class numpy:
        gaussian: Final = NumPyGaussianSampler.create
        halton: Final = NumPyHaltonSplineSampler.create

    class jax:
        gaussian: Final = JaxGaussianSampler.create
        halton: Final = JaxHaltonSplineSampler.create



from typing import Protocol
from dataclasses import dataclass

from trajax.types.array import DataType
from trajax.types.costs.boundary.common import BoundaryDistanceExtractor

from numtypes import Array, Dims


class NumPyBoundaryDistanceExtractor[StateBatchT, DistanceT](
    BoundaryDistanceExtractor[StateBatchT, DistanceT], Protocol
): ...


@dataclass(frozen=True)
class NumPyBoundaryDistance[T: int, M: int]:
    _array: Array[Dims[T, M]]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        return self._array

    @property
    def horizon(self) -> T:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> M:
        return self._array.shape[1]

    @property
    def array(self) -> Array[Dims[T, M]]:
        return self._array



from typing import Final

from trajax.trajectories.line import NumPyLineTrajectory, JaxLineTrajectory
from trajax.trajectories.waypoints import (
    NumPyWaypointsTrajectory,
    JaxWaypointsTrajectory,
)


class trajectory:
    class numpy:
        line: Final = NumPyLineTrajectory.create
        waypoints: Final = NumPyWaypointsTrajectory.create

    class jax:
        line: Final = JaxLineTrajectory.create
        waypoints: Final = JaxWaypointsTrajectory.create



from typing import TypedDict


from trajax.types import (
    Mppi,
    Trajectory,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxDynamicalModel,
    JaxSampler,
    JaxCosts,
    JaxCostFunction,
    JaxUpdateFunction,
    JaxPaddingFunction,
    JaxFilterFunction,
    JaxPathParameters,
    JaxReferencePoints,
    JaxPositionExtractor,
)
from trajax.states import (
    JaxSimpleState,
    JaxSimpleControlInputSequence,
    JaxAugmentedState,
    JaxAugmentedStateSequence,
    JaxAugmentedStateBatch,
    JaxAugmentedControlInputSequence,
    JaxAugmentedControlInputBatch,
    JaxAugmentedMppi,
    JaxSimpleControlInputBatch,
    AugmentedModel,
    extract,
)
from trajax.models import model as create_model
from trajax.samplers import sampler as create_sampler
from trajax.costs import (
    costs as create_costs,
    JaxContouringCost,
    JaxLagCost,
)
from trajax.mppi import JaxWeights
from trajax.mpcc.common import MpccMppiSetup

from numtypes import D
from jaxtyping import Float, Array as JaxArray
from deepmerge import always_merger

import jax.numpy as jnp
import jax.random as jrandom


type JaxMpccVirtualState = JaxSimpleState[D[1]]
type JaxMpccAugmentedState[S: JaxState] = JaxAugmentedState[S, JaxMpccVirtualState]

type JaxMpccVirtualStateSequence = JaxStateSequence[int, D[1]]
type JaxMpccAugmentedStateSequence[SS: JaxStateSequence] = JaxAugmentedStateSequence[
    SS, JaxMpccVirtualStateSequence
]

type JaxMpccVirtualStateBatch = JaxStateBatch[int, D[1], int]
type JaxMpccAugmentedStateBatch[SB: JaxStateBatch] = JaxAugmentedStateBatch[
    SB, JaxMpccVirtualStateBatch
]

type JaxMpccVirtualControlInputSequence = JaxSimpleControlInputSequence[int, D[1]]
type JaxMpccAugmentedControlInputSequence[CS: JaxControlInputSequence] = (
    JaxAugmentedControlInputSequence[CS, JaxMpccVirtualControlInputSequence]
)

type JaxMpccVirtualControlInputBatch = JaxControlInputBatch[int, D[1], int]
type JaxMpccAugmentedControlInputBatch[CB: JaxControlInputBatch] = (
    JaxAugmentedControlInputBatch[CB, JaxMpccVirtualControlInputBatch]
)

type JaxMpccCostFunction[
    CB: JaxControlInputBatch,
    SB: JaxStateBatch,
    C: JaxCosts,
] = JaxCostFunction[
    JaxMpccAugmentedControlInputBatch[CB], JaxMpccAugmentedStateBatch[SB], C
]


class JaxMpccWeightConfig:
    class Full(TypedDict):
        contouring: float
        lag: float
        progress: float

    class Partial(TypedDict, total=False):
        contouring: float
        lag: float
        progress: float


class JaxMpccVirtualStateConfig:
    class Full(TypedDict):
        state_limits: tuple[float, float]
        velocity_limits: tuple[float, float]
        sampling_standard_deviation: float
        sampling_seed: int
        periodic: bool

    class Partial(TypedDict, total=False):
        state_limits: tuple[float, float]
        velocity_limits: tuple[float, float]
        sampling_standard_deviation: float
        sampling_seed: int
        periodic: bool


class JaxMpccConfig:
    class Full(TypedDict):
        weights: JaxMpccWeightConfig.Full
        virtual: JaxMpccVirtualStateConfig.Full

    class Partial(TypedDict, total=False):
        weights: JaxMpccWeightConfig.Partial
        virtual: JaxMpccVirtualStateConfig.Partial


def fill_defaults(
    config: JaxMpccConfig.Partial | None,
    *,
    reference: Trajectory[JaxPathParameters, JaxReferencePoints],
) -> JaxMpccConfig.Full:
    return always_merger.merge(  # type: ignore
        JaxMpccConfig.Full(
            {
                "weights": {
                    "contouring": 20.0,
                    "lag": 10.0,
                    "progress": 500.0,
                },
                "virtual": {
                    "state_limits": (0.0, reference.path_length),
                    "velocity_limits": (0.0, 15.0),
                    "sampling_standard_deviation": 1.0,
                    "sampling_seed": 0,
                    "periodic": False,
                },
            }
        ),
        config if config is not None else JaxMpccConfig.Partial({}),
    )


class JaxMpccMppi:
    @staticmethod
    def create[
        S: JaxState,
        SS: JaxStateSequence,
        SB: JaxStateBatch,
        CS: JaxControlInputSequence,
        CB: JaxControlInputBatch,
        C: JaxCosts,
    ](
        *,
        planning_interval: int = 1,
        model: JaxDynamicalModel[S, SS, SB, CS, CB],
        sampler: JaxSampler[CS, CB],
        costs: tuple[JaxMpccCostFunction[CB, SB, C], ...],
        reference: Trajectory[JaxPathParameters, JaxReferencePoints],
        position_extractor: JaxPositionExtractor[JaxMpccAugmentedStateBatch[SB]],
        config: JaxMpccConfig.Partial | None = None,
        update_function: JaxUpdateFunction[JaxMpccAugmentedControlInputSequence[CS]]
        | None = None,
        padding_function: JaxPaddingFunction[
            JaxMpccAugmentedControlInputSequence[CS],
            JaxAugmentedControlInputSequence,
        ]
        | None = None,
        filter_function: JaxFilterFunction[JaxMpccAugmentedControlInputSequence[CS]]
        | None = None,
    ) -> MpccMppiSetup[
        Mppi[
            JaxMpccAugmentedState[S],
            JaxMpccAugmentedControlInputSequence[CS],
            JaxWeights,
        ],
        AugmentedModel[
            S,
            SS,
            SB,
            CS,
            CB,
            JaxMpccVirtualState,
            JaxMpccVirtualStateSequence,
            JaxMpccVirtualStateBatch,
            JaxMpccVirtualControlInputSequence,
            JaxMpccVirtualControlInputBatch,
            JaxMpccAugmentedState[S],
            JaxMpccAugmentedStateSequence[SS],
            JaxMpccAugmentedStateBatch[SB],
        ],
        JaxContouringCost[JaxMpccAugmentedStateBatch[SB]],
        JaxLagCost[JaxMpccAugmentedStateBatch[SB]],
    ]:
        full_config = fill_defaults(config, reference=reference)

        planner, augmented_model = JaxAugmentedMppi.create(
            planning_interval=planning_interval,
            models=(
                model,
                create_model.jax.integrator.dynamical(
                    time_step_size=model.time_step_size,
                    state_limits=full_config["virtual"]["state_limits"],
                    velocity_limits=full_config["virtual"]["velocity_limits"],
                    periodic=full_config["virtual"]["periodic"],
                ),
            ),
            samplers=(
                sampler,
                create_sampler.jax.gaussian(
                    standard_deviation=jnp.array(
                        [full_config["virtual"]["sampling_standard_deviation"]]
                    ),
                    rollout_count=sampler.rollout_count,
                    to_batch=JaxSimpleControlInputBatch.create,
                    key=jrandom.key(full_config["virtual"]["sampling_seed"]),
                ),
            ),
            cost=create_costs.jax.combined(
                contouring_cost := create_costs.jax.tracking.contouring(
                    reference=reference,
                    path_parameter_extractor=(
                        path_parameter_extractor := extract.from_virtual(
                            extract_path_parameters
                        )
                    ),
                    position_extractor=position_extractor,
                    weight=full_config["weights"]["contouring"],
                ),
                lag_cost := create_costs.jax.tracking.lag(
                    reference=reference,
                    path_parameter_extractor=path_parameter_extractor,
                    position_extractor=position_extractor,
                    weight=full_config["weights"]["lag"],
                ),
                create_costs.jax.tracking.progress(
                    path_velocity_extractor=extract.from_virtual(extract_path_velocity),
                    time_step_size=model.time_step_size,
                    weight=full_config["weights"]["progress"],
                ),
                *costs,
            ),
            state=JaxAugmentedState,
            state_sequence=JaxAugmentedStateSequence,
            state_batch=JaxAugmentedStateBatch,  # type: ignore
            input_batch=JaxAugmentedControlInputBatch,
            update_function=update_function,
            padding_function=padding_function,
            filter_function=filter_function,
        )

        return MpccMppiSetup(
            mppi=planner,  # type: ignore
            model=augmented_model,
            contouring_cost=contouring_cost,
            lag_cost=lag_cost,
        )


def extract_path_parameters(states: JaxMpccVirtualStateBatch) -> JaxPathParameters:
    return JaxPathParameters(states.array[:, 0, :])


def extract_path_velocity(
    inputs: JaxMpccVirtualControlInputBatch,
) -> Float[JaxArray, "T M"]:
    return inputs.array[:, 0, :]



from typing import Self
from dataclasses import dataclass

from trajax.types import jaxtyped, JaxObstacleSimulator
from trajax.obstacles.accelerated import JaxObstacleStatesForTimeStep

from jaxtyping import Array as JaxArray, Float, Scalar

import jax
import jax.numpy as jnp


@dataclass(kw_only=True)
class JaxDynamicObstacleSimulator[K: int](
    JaxObstacleSimulator[JaxObstacleStatesForTimeStep[K]]
):
    last: JaxObstacleStatesForTimeStep[K]
    velocities: Float[JaxArray, "K 2"]

    time_step: Scalar | None = None

    @staticmethod
    def create[K_: int = int](
        *,
        positions: Float[JaxArray, "K 2"],
        velocities: Float[JaxArray, "K 2"],
        obstacle_count: K_ | None = None,
    ) -> "JaxDynamicObstacleSimulator[K_]":
        headings = headings_from(velocities)
        count = positions.shape[0]

        assert obstacle_count is None or obstacle_count == count, (
            f"Expected {obstacle_count} obstacles, but got {count}."
        )

        return JaxDynamicObstacleSimulator(
            last=JaxObstacleStatesForTimeStep.create(
                x=positions[:, 0], y=positions[:, 1], heading=headings
            ),
            velocities=velocities,
        )

    def with_time_step_size(self, time_step_size: float) -> Self:
        return self.__class__(
            last=self.last,
            velocities=self.velocities,
            time_step=jnp.asarray(time_step_size),
        )

    def step(self) -> JaxObstacleStatesForTimeStep[K]:
        assert self.time_step is not None, (
            "Time step must be set to advance obstacle states."
        )

        x, y = step_obstacles(
            x=self.last.x_array,
            y=self.last.y_array,
            velocities=self.velocities,
            time_step=self.time_step,
        )

        self.last = JaxObstacleStatesForTimeStep.create(
            x=x, y=y, heading=self.last.heading_array
        )

        return self.last

    @property
    def obstacle_count(self) -> int:
        return self.last.count


@jax.jit
@jaxtyped
def step_obstacles(
    *,
    x: Float[JaxArray, "K"],
    y: Float[JaxArray, "K"],
    velocities: Float[JaxArray, "K 2"],
    time_step: Scalar,
) -> tuple[Float[JaxArray, "K"], Float[JaxArray, "K"]]:
    new_x = x + velocities[:, 0] * time_step
    new_y = y + velocities[:, 1] * time_step
    return new_x, new_y


@jax.jit
@jaxtyped
def headings_from(velocities: Float[JaxArray, "K 2"]) -> Float[JaxArray, "K"]:
    speed = jnp.linalg.norm(velocities, axis=1)
    return jnp.where(speed > 1e-6, jnp.arctan2(velocities[:, 1], velocities[:, 0]), 0.0)



from typing import Final, TypeAlias

from trajax.namespace import types as _types, classes as _classes
from trajax.factory import (
    mppi as _mppi,
    update as _update,
    padding as _padding,
    filters as _filters,
)
from trajax.costs import (
    costs as _costs,
    distance as _distance,
    boundary as _boundary,
    risk as _risk,
)
from trajax.models import model as _model
from trajax.obstacles import obstacles as _obstacles
from trajax.predictors import predictor as _predictor, propagator as _propagator
from trajax.samplers import sampler as _sampler
from trajax.trajectories import trajectory as _trajectory
from trajax.states import extract as _extract

types: TypeAlias = _types.numpy
classes: Final = _classes.numpy
mppi: Final = _mppi.numpy
update: Final = _update.numpy
padding: Final = _padding.numpy
filters: Final = _filters.numpy
costs: Final = _costs.numpy
distance: Final = _distance.numpy
boundary: Final = _boundary.numpy
risk: Final = _risk.numpy
model: Final = _model.numpy
obstacles: Final = _obstacles.numpy
predictor: Final = _predictor.numpy
propagator: Final = _propagator.numpy
sampler: Final = _sampler.numpy
trajectory: Final = _trajectory.numpy
extract: Final = _extract



from dataclasses import dataclass

from trajax.types import (
    jaxtyped,
    DistanceExtractor,
    JaxPositionExtractor,
    JaxHeadingExtractor,
)
from trajax.obstacles import JaxSampledObstacleStates
from trajax.costs.collision import JaxDistance
from trajax.costs.distance.sat.common import ConvexPolygon
from trajax.costs.distance.accelerated import replace_missing

from jaxtyping import Array as JaxArray, Float, Scalar, Bool

import jax
import jax.numpy as jnp


type BoolScalar = Bool[JaxArray, ""]
type Vector2d = Float[JaxArray, "2"]


@dataclass(frozen=True)
class JaxSatDistanceExtractor[StateT](
    DistanceExtractor[StateT, JaxSampledObstacleStates, JaxDistance]
):
    """
    Computes the signed distances between the ego polygon and obstacle polygons
    using the Separating Axis Theorem (SAT).

    Positive distance means separated, negative means penetrating.
    """

    ego_vertices: Float[JaxArray, "P_ego 2"]
    obstacle_vertices: Float[JaxArray, "P_obs 2"]
    positions_from: JaxPositionExtractor[StateT]
    headings_from: JaxHeadingExtractor[StateT]

    @staticmethod
    def create[S](
        *,
        ego: ConvexPolygon,
        obstacle: ConvexPolygon,
        position_extractor: JaxPositionExtractor[S],
        heading_extractor: JaxHeadingExtractor[S],
    ) -> "JaxSatDistanceExtractor[S]":
        return JaxSatDistanceExtractor(
            ego_vertices=jnp.asarray(ego.vertices),
            obstacle_vertices=jnp.asarray(obstacle.vertices),
            positions_from=position_extractor,
            headings_from=heading_extractor,
        )

    def __call__[T: int, N: int](
        self,
        *,
        states: StateT,
        obstacle_states: JaxSampledObstacleStates[T, int, N],
    ) -> JaxDistance[T, int, int, N]:
        ego_positions = self.positions_from(states)
        ego_headings = self.headings_from(states)

        if self._no_obstacles_exist(obstacle_states):
            T = ego_positions.horizon
            M = ego_positions.rollout_count
            V = 1
            N = obstacle_states.sample_count
            return JaxDistance(jnp.full((T, V, M, N), jnp.inf))

        return JaxDistance(
            compute_sat_distances(
                ego_x=ego_positions.x_array,
                ego_y=ego_positions.y_array,
                ego_heading=ego_headings.heading,
                ego_vertices=self.ego_vertices,
                obstacle_x=obstacle_states.x_array,
                obstacle_y=obstacle_states.y_array,
                obstacle_heading=obstacle_states.heading_array,
                obstacle_vertices=self.obstacle_vertices,
            )
        )

    def _no_obstacles_exist(self, states: JaxSampledObstacleStates) -> bool:
        return states.x_array.shape[1] == 0


@jax.jit
@jaxtyped
def compute_sat_distances(
    *,
    ego_x: Float[JaxArray, "T M"],
    ego_y: Float[JaxArray, "T M"],
    ego_heading: Float[JaxArray, "T M"],
    ego_vertices: Float[JaxArray, "P_ego 2"],
    obstacle_x: Float[JaxArray, "T K N"],
    obstacle_y: Float[JaxArray, "T K N"],
    obstacle_heading: Float[JaxArray, "T K N"],
    obstacle_vertices: Float[JaxArray, "P_obs 2"],
) -> Float[JaxArray, "T V M N"]:
    obstacle_x, obstacle_y, obstacle_heading = replace_missing(
        x=obstacle_x, y=obstacle_y, heading=obstacle_heading
    )

    def compute_single_distance(
        ego_x_single: Scalar,
        ego_y_single: Scalar,
        ego_heading_single: Scalar,
        obs_x_single: Scalar,
        obs_y_single: Scalar,
        obs_heading_single: Scalar,
    ) -> Scalar:
        is_inf = jnp.isinf(obs_x_single) | jnp.isinf(obs_y_single)

        ego_transformed = transform_polygon_jax(
            ego_vertices,
            position=jnp.array([ego_x_single, ego_y_single]),
            heading=ego_heading_single,
        )
        obs_transformed = transform_polygon_jax(
            obstacle_vertices,
            position=jnp.array([obs_x_single, obs_y_single]),
            heading=obs_heading_single,
        )

        dist = sat_distance_jax(ego_transformed, obs_transformed)
        return jnp.where(is_inf, jnp.inf, dist)  # type: ignore

    def compute_for_ego_and_sample(
        ego_x_tm: Scalar,
        ego_y_tm: Scalar,
        ego_heading_tm: Scalar,
        obs_x_kn: Float[JaxArray, "K"],
        obs_y_kn: Float[JaxArray, "K"],
        obs_heading_kn: Float[JaxArray, "K"],
    ) -> Scalar:
        distances_k = jax.vmap(
            lambda ox, oy, oh: compute_single_distance(
                ego_x_tm, ego_y_tm, ego_heading_tm, ox, oy, oh
            )
        )(obs_x_kn, obs_y_kn, obs_heading_kn)
        return jnp.min(distances_k)

    def compute_for_timestep(
        ego_x_t: Float[JaxArray, "M"],
        ego_y_t: Float[JaxArray, "M"],
        ego_heading_t: Float[JaxArray, "M"],
        obs_x_t: Float[JaxArray, "K N"],
        obs_y_t: Float[JaxArray, "K N"],
        obs_heading_t: Float[JaxArray, "K N"],
    ) -> Float[JaxArray, "M N"]:
        def compute_for_ego(ex, ey, eh):
            return jax.vmap(
                lambda ox_n, oy_n, oh_n: compute_for_ego_and_sample(
                    ex, ey, eh, ox_n, oy_n, oh_n
                ),
                in_axes=(1, 1, 1),
            )(obs_x_t, obs_y_t, obs_heading_t)

        return jax.vmap(compute_for_ego)(ego_x_t, ego_y_t, ego_heading_t)

    distances = jax.vmap(compute_for_timestep)(
        ego_x,
        ego_y,
        ego_heading,
        obstacle_x,
        obstacle_y,
        obstacle_heading,
    )

    return distances[:, None, :, :]


@jax.jit
@jaxtyped
def transform_polygon_jax(
    local_vertices: Float[JaxArray, "P 2"],
    position: Float[JaxArray, "2"],
    heading: Scalar,
) -> Float[JaxArray, "P 2"]:
    cos_h = jnp.cos(heading)
    sin_h = jnp.sin(heading)
    rotation = jnp.array([[cos_h, -sin_h], [sin_h, cos_h]])
    return (local_vertices @ rotation.T) + position


@jax.jit
@jaxtyped
def get_edge_normals_jax(
    vertices: Float[JaxArray, "P 2"],
) -> Float[JaxArray, "P 2"]:
    edges = jnp.roll(vertices, -1, axis=0) - vertices
    normals = jnp.stack([-edges[:, 1], edges[:, 0]], axis=-1)
    norms = jnp.linalg.norm(normals, axis=-1, keepdims=True)
    return normals / jnp.where(norms > 1e-10, norms, 1.0)  # type: ignore


@jax.jit
@jaxtyped
def project_polygon_jax(
    vertices: Float[JaxArray, "P 2"], axis: Vector2d
) -> tuple[Scalar, Scalar]:
    projections = vertices @ axis
    return projections.min(), projections.max()


@jax.jit
@jaxtyped
def sat_distance_jax(
    poly_a: Float[JaxArray, "P_a 2"],
    poly_b: Float[JaxArray, "P_b 2"],
) -> Scalar:
    normals_a = get_edge_normals_jax(poly_a)
    normals_b = get_edge_normals_jax(poly_b)
    axes = jnp.vstack([normals_a, normals_b])

    def check_axis(axis: Vector2d) -> tuple[BoolScalar, Scalar, Scalar]:
        min_a, max_a = project_polygon_jax(poly_a, axis)
        min_b, max_b = project_polygon_jax(poly_b, axis)

        gap_ab = min_b - max_a
        gap_ba = min_a - max_b
        gap = jnp.maximum(gap_ab, gap_ba)

        overlap = jnp.minimum(max_a - min_b, max_b - min_a)

        is_separated = gap > 0
        return is_separated, gap, -overlap

    is_separated_arr, gaps, penetrations = jax.vmap(check_axis)(axes)

    any_separated = jnp.any(is_separated_arr)
    min_separation = jnp.where(is_separated_arr, gaps, jnp.inf).min()
    max_penetration = jnp.where(~is_separated_arr, penetrations, -jnp.inf).max()

    return jnp.where(any_separated, min_separation, max_penetration)



from typing import cast, Self, overload, Sequence, Protocol
from dataclasses import dataclass

from trajax.types import (
    DataType,
    jaxtyped,
    ControlInputSequence,
    JaxState,
    JaxStateSequence,
    JaxStateBatch,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxCosts,
)

from jaxtyping import Array as JaxArray, Float
from numtypes import Array, Dims, D

import numpy as np
import jax.numpy as jnp


class JaxControlInputSequenceLike[T: int, D_u: int](
    ControlInputSequence[T, D_u], Protocol
):
    @property
    def array(self) -> Float[JaxArray, "T D_u"]:
        """Returns the underlying JAX array representing the control input sequence."""
        ...


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleState[D_x: int](JaxState[D_x]):
    _array: Float[JaxArray, "D_x"]

    @staticmethod
    def zeroes[D_x_: int](*, dimension: D_x_) -> "JaxSimpleState[D_x_]":
        """Creates a zeroed simple state for the given dimension."""
        return JaxSimpleState(jnp.zeros((dimension,)))

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D_x]]:
        return np.asarray(self.array)

    @property
    def dimension(self) -> D_x:
        return cast(D_x, self.array.shape[0])

    @property
    def array(self) -> Float[JaxArray, "D_x"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleStateSequence[T: int, D_x: int](JaxStateSequence[T, D_x]):
    _array: Float[JaxArray, "T D_x"]

    @staticmethod
    def of_states[T_: int, D_x_: int](
        states: Sequence[JaxSimpleState[D_x_]], *, horizon: T_ | None = None
    ) -> "JaxSimpleStateSequence[T_, D_x_]":
        """Creates a simple state sequence from a sequence of simple states."""
        assert len(states) > 0, "States sequence must not be empty."

        horizon = horizon if horizon is not None else cast(T_, len(states))
        dimension = states[0].dimension
        array = jnp.stack([state.array for state in states], axis=0)

        assert array.shape == (horizon, dimension), (
            f"Expected array shape {(horizon, dimension)}, but got {array.shape}"
        )

        return JaxSimpleStateSequence(array)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_x]]:
        return np.asarray(self.array)

    def batched(self) -> "JaxSimpleStateBatch[T, D_x, D[1]]":
        return JaxSimpleStateBatch.wrap(self.array[..., jnp.newaxis])

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def dimension(self) -> D_x:
        return cast(D_x, self.array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T D_x"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleStateBatch[T: int, D_x: int, M: int](JaxStateBatch[T, D_x, M]):
    _array: Float[JaxArray, "T D_x M"]

    @staticmethod
    def wrap(array: Float[JaxArray, "T D_x M"]) -> "JaxSimpleStateBatch[T, D_x, M]":
        """Creates a JAX simple state batch from the given array."""
        return JaxSimpleStateBatch(array)

    @staticmethod
    def of_states[D_x_: int, T_: int = int](
        states: Sequence[JaxState[D_x_]], *, horizon: T_ | None = None
    ) -> "JaxSimpleStateBatch[T_, D_x_, int]":
        """Creates a simple state batch from a sequence of simple states."""
        assert len(states) > 0, "States sequence must not be empty."

        horizon = horizon if horizon is not None else cast(T_, len(states))
        dimension = states[0].dimension
        array = jnp.stack([state.array for state in states], axis=0)[:, :, jnp.newaxis]

        assert array.shape == (horizon, dimension, 1), (
            f"Expected array shape {(horizon, dimension, 1)}, but got {array.shape}"
        )

        return JaxSimpleStateBatch(array)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_x, M]]:
        return np.asarray(self.array)

    def rollout(self, index: int) -> JaxSimpleStateSequence[T, D_x]:
        return JaxSimpleStateSequence(self.array[..., index])

    def at(self, *, time_step: int, rollout: int) -> JaxSimpleState[D_x]:
        return JaxSimpleState(self.array[time_step, :, rollout])

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def dimension(self) -> D_x:
        return cast(D_x, self.array.shape[1])

    @property
    def rollout_count(self) -> M:
        return cast(M, self.array.shape[2])

    @property
    def array(self) -> Float[JaxArray, "T D_x M"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleControlInputSequence[T: int, D_u: int](JaxControlInputSequence[T, D_u]):
    _array: Float[JaxArray, "T D_u"]

    @staticmethod
    def constant[T_: int, D_u_: int](
        value: Float[JaxArray, "D_u"] | Array[Dims[D_u_]], *, horizon: T_
    ) -> "JaxSimpleControlInputSequence[T_, D_u_]":
        """Creates a constant control input sequence for the given horizon and dimension."""
        return JaxSimpleControlInputSequence(
            jnp.tile(jnp.asarray(value)[jnp.newaxis, :], (horizon, 1))
        )

    @staticmethod
    def zeroes[T_: int, D_u_: int](
        horizon: T_, dimension: D_u_
    ) -> "JaxSimpleControlInputSequence[T_, D_u_]":
        """Creates a zeroed control input sequence for the given horizon."""
        return JaxSimpleControlInputSequence.constant(
            jnp.zeros((dimension,)), horizon=horizon
        )

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_u]]:
        return np.asarray(self.array)

    @overload
    def similar(self, *, array: Float[JaxArray, "T D_u"]) -> Self: ...

    @overload
    def similar[L: int](
        self, *, array: Float[JaxArray, "L D_u"], length: L
    ) -> "JaxSimpleControlInputSequence[L, D_u]": ...

    def similar[L: int](
        self, *, array: Float[JaxArray, "L D_u"], length: L | None = None
    ) -> "Self | JaxSimpleControlInputSequence[L, D_u]":
        length = length if length is not None else cast(L, array.shape[0])

        assert array.shape[0] == length, (
            f"Expected array with time horizon {length}, but got {array.shape[0]}"
        )

        return self.__class__(array)

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def dimension(self) -> D_u:
        return cast(D_u, self.array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T D_u"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleControlInputBatch[T: int, D_u: int, M: int](
    JaxControlInputBatch[T, D_u, M]
):
    _array: Float[JaxArray, "T D_u M"]

    @staticmethod
    def create(*, array: Float[JaxArray, "T D_u M"]) -> "JaxSimpleControlInputBatch":
        """Factory method to create a JAX simple control input batch from an array."""
        return JaxSimpleControlInputBatch(array)

    @staticmethod
    def zero[T_: int, D_u_: int, M_: int](
        *, horizon: T_, dimension: D_u_, rollout_count: M_ = 1
    ) -> "JaxSimpleControlInputBatch[T_, D_u_, M_]":
        """Creates a zeroed control input batch for the given horizon and rollout count."""
        return JaxSimpleControlInputBatch(
            jnp.zeros((horizon, dimension, rollout_count))
        )

    @staticmethod
    def of[T_: int, D_u_: int](
        sequence: JaxControlInputSequenceLike[T_, D_u_],
    ) -> "JaxSimpleControlInputBatch[T_, D_u_, D[1]]":
        """Creates a JAX simple control input batch from a single control input sequence."""
        array = sequence.array[..., jnp.newaxis]

        assert array.shape == (sequence.horizon, sequence.dimension, 1), (
            f"Expected array shape {(sequence.horizon, sequence.dimension, 1)}, "
            f"but got {array.shape}"
        )

        return JaxSimpleControlInputBatch(array)

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_u, M]]:
        return np.asarray(self.array)

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def dimension(self) -> D_u:
        return cast(D_u, self.array.shape[1])

    @property
    def rollout_count(self) -> M:
        return cast(M, self.array.shape[2])

    @property
    def array(self) -> Float[JaxArray, "T D_u M"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class JaxSimpleCosts[T: int, M: int](JaxCosts[T, M]):
    _array: Float[JaxArray, "T M"]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, M]]:
        return np.asarray(self.array)

    def similar(self, *, array: Float[JaxArray, "T M"]) -> Self:
        return self.__class__(array)

    def zero(self) -> Self:
        return self.__class__(jnp.zeros_like(self.array))

    @property
    def horizon(self) -> T:
        return cast(T, self.array.shape[0])

    @property
    def rollout_count(self) -> M:
        return cast(M, self.array.shape[1])

    @property
    def array(self) -> Float[JaxArray, "T M"]:
        return self._array



from typing import Protocol

from trajax.types.array import DataType

from numtypes import Array, Dims


class ObstaclePositionsForTimeStep[D_p: int, K: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D_p, K]]:
        """Returns the obstacle positions for a single time step as a NumPy array."""
        ...

    @property
    def dimension(self) -> D_p:
        """The dimension of the obstacle positions."""
        ...

    @property
    def count(self) -> K:
        """The number of obstacles."""
        ...


class ObstaclePositions[T: int, D_p: int, K: int](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_p, K]]:
        """Returns the obstacle positions as a NumPy array."""
        ...

    @property
    def horizon(self) -> T:
        """The time horizon of the obstacle positions."""
        ...

    @property
    def dimension(self) -> D_p:
        """The dimension of the obstacle positions."""
        ...

    @property
    def count(self) -> K:
        """The number of obstacles."""
        ...


class ObstaclePositionExtractor[
    ObstacleStatesForTimeStepT,
    ObstacleStatesT,
    PositionsForTimeStepT,
    PositionsT,
](Protocol):
    def of_states_for_time_step(
        self, states: ObstacleStatesForTimeStepT, /
    ) -> PositionsForTimeStepT:
        """Extracts the positions from the given obstacle states for a single time step."""
        ...

    def of_states(self, states: ObstacleStatesT, /) -> PositionsT:
        """Extracts the positions from the given obstacle states."""
        ...



from typing import Mapping, Any, overload
from dataclasses import dataclass

from trajax.types import SimulationData, Metric

from trajax.collectors import CollectorRegistry


@dataclass
class MetricRegistry:
    metrics_by_name: Mapping[str, Metric]
    computed_metrics_by_name: dict[str, Any]
    collectors: CollectorRegistry

    @staticmethod
    def of(*metrics: Metric, collectors: CollectorRegistry) -> "MetricRegistry":
        return MetricRegistry(
            metrics_by_name={metric.name: metric for metric in metrics},
            computed_metrics_by_name={},
            collectors=collectors,
        )

    def __post_init__(self) -> None:
        self.collectors.on_modified(self._modified)

    @overload
    def get[R](self, metric: Metric[R]) -> R:
        """Retrieves results for the specified metric from the registry."""
        ...

    @overload
    def get(self, metric: str) -> Any:
        """Retrieves results for the specified metric from the registry."""
        ...

    def get[R](self, metric: Metric[R] | str) -> R | Any:
        if (
            name := metric if isinstance(metric, str) else metric.name
        ) not in self.computed_metrics_by_name:
            self.computed_metrics_by_name[name] = self._compute_metric(name)

        return self.computed_metrics_by_name[name]

    @property
    def data(self) -> SimulationData:
        """Retrieves the collected simulation data."""
        return self.collectors.data

    def _modified(self) -> None:
        self.computed_metrics_by_name = {}

    def _compute_metric(self, name: str) -> Any:
        found_metric = self.metrics_by_name.get(name)

        assert found_metric is not None, (
            f"Metric with name '{name}' not found in registry."
        )

        return found_metric.compute(self.data)



from typing import Protocol

from trajax.types.trajectories import NumPyPathParameters, NumPyPositions, NumPyHeadings
from trajax.types.costs.common import PositionExtractor

from numtypes import Array, Dim2


class NumPyPathParameterExtractor[StateBatchT](Protocol):
    def __call__(self, states: StateBatchT, /) -> NumPyPathParameters:
        """Extracts path parameters from a batch of states."""
        ...


class NumPyPathVelocityExtractor[InputBatchT](Protocol):
    def __call__(self, inputs: InputBatchT, /) -> Array[Dim2]:
        """Extracts path velocities from a batch of control inputs."""
        ...


class NumPyPositionExtractor[StateBatchT](
    PositionExtractor[StateBatchT, NumPyPositions], Protocol
): ...


class NumPyHeadingExtractor[StateBatchT](Protocol):
    def __call__(self, states: StateBatchT, /) -> NumPyHeadings:
        """Extracts heading angles from a batch of states."""
        ...



from typing import Final

from trajax.metrics.registry import MetricRegistry
from trajax.metrics.collision import CollisionMetric
from trajax.metrics.error import MpccErrorMetric
from trajax.metrics.task import TaskCompletionMetric
from trajax.metrics.constraint import ConstraintViolationMetric
from trajax.metrics.comfort import ComfortMetric


class metrics:
    registry: Final = MetricRegistry.of
    collision: Final = CollisionMetric.create
    mpcc_error: Final = MpccErrorMetric.create
    task_completion: Final = TaskCompletionMetric.create
    constraint_violation: Final = ConstraintViolationMetric.create
    comfort: Final = ComfortMetric.create



from typing import NamedTuple, overload
from dataclasses import dataclass

from trajax.types import (
    DistanceExtractor,
    NumPyHeadings,
    NumPyPositions,
    NumPyPositionExtractor,
    NumPyHeadingExtractor,
)
from trajax.obstacles import NumPySampledObstacleStates
from trajax.costs.collision import NumPyDistance
from trajax.costs.distance.basic import replace_missing
from trajax.costs.distance.sat.common import ConvexPolygon, VerticesArray

from numtypes import Array, BoolArray, Dims, D, shape_of

import numpy as np

type V = D[1]
"""Number of ego polygons (fixed to 1 for now)."""

type C = D[1]
"""Number of obstacle polygons (fixed to 1 for now)."""


class MaskedObstacleStates[T: int, K: int, N: int](NamedTuple):
    x: Array[Dims[T, K, N]]
    y: Array[Dims[T, K, N]]
    heading: Array[Dims[T, K, N]]
    is_valid: BoolArray[Dims[T, K, N]]


@dataclass(frozen=True)
class NumPySatDistanceExtractor[StateT](
    DistanceExtractor[StateT, NumPySampledObstacleStates, NumPyDistance]
):
    """
    Computes the signed distances between the ego polygon and obstacle polygons
    using the Separating Axis Theorem (SAT).

    Positive distance means separated, negative means penetrating.
    """

    ego: ConvexPolygon
    obstacle: ConvexPolygon
    positions_from: NumPyPositionExtractor[StateT]
    headings_from: NumPyHeadingExtractor[StateT]

    @staticmethod
    def create[S](
        *,
        ego: ConvexPolygon,
        obstacle: ConvexPolygon,
        position_extractor: NumPyPositionExtractor[S],
        heading_extractor: NumPyHeadingExtractor[S],
    ) -> "NumPySatDistanceExtractor[S]":
        return NumPySatDistanceExtractor(
            ego=ego,
            obstacle=obstacle,
            positions_from=position_extractor,
            headings_from=heading_extractor,
        )

    def __call__[T: int, N: int, M: int = int, K: int = int](
        self,
        *,
        states: StateT,
        obstacle_states: NumPySampledObstacleStates[T, K, N],
    ) -> NumPyDistance[T, V, M, N]:
        return NumPyDistance(
            compute_sat_distances(
                ego_positions=self.positions_from(states),
                ego_headings=self.headings_from(states),
                ego_vertices=self.ego.vertices,
                obstacle_states=replace_missing(obstacle_states),
                obstacle_vertices=self.obstacle.vertices,
            )
        )


def compute_sat_distances[T: int, M: int, K: int, N: int](
    *,
    ego_positions: NumPyPositions[T, M],
    ego_headings: NumPyHeadings[T, M],
    ego_vertices: VerticesArray,
    obstacle_states: NumPySampledObstacleStates[T, K, N],
    obstacle_vertices: VerticesArray,
) -> Array[Dims[T, V, M, N]]:
    ego_x = ego_positions.x()
    ego_y = ego_positions.y()
    ego_heading = ego_headings.heading

    obstacle_x, obstacle_y, obstacle_heading, is_valid = mask_valid_obstacle_states(
        obstacle_states=obstacle_states
    )

    T, M = ego_x.shape
    K = obstacle_x.shape[1]
    N = obstacle_x.shape[2]
    V = 1

    if K == 0:
        distances = np.full((T, V, M, N), np.inf)
        assert shape_of(distances, matches=(T, V, M, N), name="distances")
        return distances

    ego_polygons = transform_polygons_batched(
        ego_vertices,
        x=ego_x,
        y=ego_y,
        heading=ego_heading,
    )

    obstacle_polygons = transform_polygons_batched(
        obstacle_vertices,
        x=obstacle_x,
        y=obstacle_y,
        heading=obstacle_heading,
    )

    distances = sat_distances_batched(
        ego_polygons=ego_polygons,
        obstacle_polygons=obstacle_polygons,
        is_valid=is_valid,
    )

    min_over_obstacles = np.min(distances, axis=2, keepdims=True)
    result = np.moveaxis(min_over_obstacles, 2, 1)

    return result


def mask_valid_obstacle_states[T: int, K: int, N: int](
    *,
    obstacle_states: NumPySampledObstacleStates[T, K, N],
) -> MaskedObstacleStates[T, K, N]:
    T, K, N = (
        obstacle_states.horizon,
        obstacle_states.count,
        obstacle_states.sample_count,
    )

    is_valid = ~(np.isinf(obstacle_states.x()) | np.isinf(obstacle_states.y()))
    x = np.where(is_valid, obstacle_states.x(), 0.0)
    y = np.where(is_valid, obstacle_states.y(), 0.0)
    heading = np.where(is_valid, obstacle_states.heading(), 0.0)

    assert shape_of(x, matches=(T, K, N), name="x")
    assert shape_of(y, matches=(T, K, N), name="y")
    assert shape_of(heading, matches=(T, K, N), name="heading")
    assert shape_of(is_valid, matches=(T, K, N), name="is_valid")

    return MaskedObstacleStates(x=x, y=y, heading=heading, is_valid=is_valid)


@overload
def transform_polygons_batched[T: int, M: int, P: int = int](
    local_vertices: VerticesArray,
    *,
    x: Array[Dims[T, M]],
    y: Array[Dims[T, M]],
    heading: Array[Dims[T, M]],
) -> Array[Dims[T, M, P, D[2]]]: ...


@overload
def transform_polygons_batched[T: int, K: int, N: int, P: int = int](
    local_vertices: VerticesArray,
    *,
    x: Array[Dims[T, K, N]],
    y: Array[Dims[T, K, N]],
    heading: Array[Dims[T, K, N]],
) -> Array[Dims[T, K, N, P, D[2]]]: ...


def transform_polygons_batched[T: int, M: int, K: int, N: int, P: int = int](
    local_vertices: VerticesArray,
    *,
    x: Array[Dims[T, M]] | Array[Dims[T, K, N]],
    y: Array[Dims[T, M]] | Array[Dims[T, K, N]],
    heading: Array[Dims[T, M]] | Array[Dims[T, K, N]],
) -> Array[Dims[T, M, P, D[2]]] | Array[Dims[T, K, N, P, D[2]]]:
    P = local_vertices.shape[0]
    spatial_shape = x.shape

    rotation_matrices = create_rotation_matrices(heading)
    local_expanded = local_vertices.reshape((1,) * len(spatial_shape) + (P, 2))
    rotated = np.einsum("...pj,...ij->...pi", local_expanded, rotation_matrices)

    positions = np.stack([x, y], axis=-1)
    positions_expanded = positions[..., np.newaxis, :]

    return rotated + positions_expanded


@overload
def create_rotation_matrices[T: int, M: int](
    heading: Array[Dims[T, M]],
) -> Array[Dims[T, M, D[2], D[2]]]: ...


@overload
def create_rotation_matrices[T: int, K: int, N: int](
    heading: Array[Dims[T, K, N]],
) -> Array[Dims[T, K, N, D[2], D[2]]]: ...


def create_rotation_matrices[T: int, M: int, K: int, N: int](
    heading: Array[Dims[T, M]] | Array[Dims[T, K, N]],
) -> Array[Dims[T, M, D[2], D[2]]] | Array[Dims[T, K, N, D[2], D[2]]]:
    cos_h = np.cos(heading)
    sin_h = np.sin(heading)
    return np.stack(
        [
            np.stack([cos_h, -sin_h], axis=-1),
            np.stack([sin_h, cos_h], axis=-1),
        ],
        axis=-2,
    )


def sat_distances_batched[
    T: int,
    M: int,
    K: int,
    N: int,
    P_ego: int = int,
    P_obs: int = int,
](
    *,
    ego_polygons: Array[Dims[T, M, P_ego, D[2]]],
    obstacle_polygons: Array[Dims[T, K, N, P_obs, D[2]]],
    is_valid: BoolArray[Dims[T, K, N]],
) -> Array[Dims[T, M, K, N]]:
    all_axes = gather_separating_axes(ego_polygons, obstacle_polygons)

    ego_expanded = ego_polygons[:, :, np.newaxis, np.newaxis, :, :]
    obs_expanded = obstacle_polygons[:, np.newaxis, :, :, :, :]

    ego_min, ego_max = project_polygons_onto_axes(ego_expanded, all_axes)
    obs_min, obs_max = project_polygons_onto_axes(obs_expanded, all_axes)

    distances = compute_signed_distances_from_projections(
        ego_min, ego_max, obs_min, obs_max
    )

    is_invalid_expanded = ~is_valid[:, np.newaxis, :, :]
    return np.where(is_invalid_expanded, np.inf, distances)


def gather_separating_axes[
    T: int,
    M: int,
    K: int,
    N: int,
    P_ego: int = int,
    P_obs: int = int,
    L_axes: int = int,
](
    ego_polygons: Array[Dims[T, M, P_ego, D[2]]],
    obstacle_polygons: Array[Dims[T, K, N, P_obs, D[2]]],
) -> Array[Dims[T, M, K, N, L_axes, D[2]]]:
    P_ego = ego_polygons.shape[2]
    P_obs = obstacle_polygons.shape[3]

    ego_normals = get_edge_normals_batched(ego_polygons)
    obs_normals = get_edge_normals_batched(obstacle_polygons)

    ego_normals_expanded = ego_normals[:, :, np.newaxis, np.newaxis, :, :]
    obs_normals_expanded = obs_normals[:, np.newaxis, :, :, :, :]

    T, M = ego_polygons.shape[:2]
    K, N = obstacle_polygons.shape[1:3]

    ego_normals_full = np.broadcast_to(ego_normals_expanded, (T, M, K, N, P_ego, 2))
    obs_normals_full = np.broadcast_to(obs_normals_expanded, (T, M, K, N, P_obs, 2))

    return np.concatenate([ego_normals_full, obs_normals_full], axis=-2)


@overload
def project_polygons_onto_axes[
    T: int,
    M: int,
    K: int,
    N: int,
    P: int = int,
    L_axes: int = int,
](
    polygons: Array[Dims[T, M, D[1], D[1], P, D[2]]],
    axes: Array[Dims[T, M, K, N, L_axes, D[2]]],
) -> tuple[Array[Dims[T, M, K, N, L_axes]], Array[Dims[T, M, K, N, L_axes]]]: ...


@overload
def project_polygons_onto_axes[
    T: int,
    M: int,
    K: int,
    N: int,
    P: int = int,
    L_axes: int = int,
](
    polygons: Array[Dims[T, D[1], K, N, P, D[2]]],
    axes: Array[Dims[T, M, K, N, L_axes, D[2]]],
) -> tuple[Array[Dims[T, M, K, N, L_axes]], Array[Dims[T, M, K, N, L_axes]]]: ...


def project_polygons_onto_axes[
    T: int,
    M: int,
    K: int,
    N: int,
    P: int = int,
    L_axes: int = int,
](
    polygons: Array[Dims[T, M, D[1], D[1], P, D[2]]]
    | Array[Dims[T, D[1], K, N, P, D[2]]],
    axes: Array[Dims[T, M, K, N, L_axes, D[2]]],
) -> tuple[Array[Dims[T, M, K, N, L_axes]], Array[Dims[T, M, K, N, L_axes]]]:
    projections = np.einsum("...pi,...ai->...ap", polygons, axes)
    return projections.min(axis=-1), projections.max(axis=-1)


def compute_signed_distances_from_projections[
    T: int,
    M: int,
    K: int,
    N: int,
    L_axes: int = int,
](
    ego_min: Array[Dims[T, M, K, N, L_axes]],
    ego_max: Array[Dims[T, M, K, N, L_axes]],
    obs_min: Array[Dims[T, M, K, N, L_axes]],
    obs_max: Array[Dims[T, M, K, N, L_axes]],
) -> Array[Dims[T, M, K, N]]:
    gap_ab = obs_min - ego_max
    gap_ba = ego_min - obs_max
    gaps = np.maximum(gap_ab, gap_ba)

    overlap = np.minimum(ego_max - obs_min, obs_max - ego_min)
    penetration = -overlap

    is_separated = gaps > 0
    any_separated = np.any(is_separated, axis=-1)

    separation_values = np.where(is_separated, gaps, np.inf)
    min_separation = np.min(separation_values, axis=-1)

    penetration_values = np.where(~is_separated, penetration, -np.inf)
    max_penetration = np.max(penetration_values, axis=-1)

    return np.where(any_separated, min_separation, max_penetration)


@overload
def get_edge_normals_batched[T: int, M: int, P_ego: int](
    vertices: Array[Dims[T, M, P_ego, D[2]]],
) -> Array[Dims[T, M, P_ego, D[2]]]: ...


@overload
def get_edge_normals_batched[T: int, K: int, N: int, P_obs: int](
    vertices: Array[Dims[T, K, N, P_obs, D[2]]],
) -> Array[Dims[T, K, N, P_obs, D[2]]]: ...


def get_edge_normals_batched[T: int, M: int, K: int, N: int, P: int](
    vertices: Array[Dims[T, M, P, D[2]]] | Array[Dims[T, K, N, P, D[2]]],
) -> Array[Dims[T, M, P, D[2]]] | Array[Dims[T, K, N, P, D[2]]]:
    edges = np.roll(vertices, -1, axis=-2) - vertices
    normals = np.stack([-edges[..., 1], edges[..., 0]], axis=-1)
    norms = np.linalg.norm(normals, axis=-1, keepdims=True)
    return normals / np.where(norms > 1e-10, norms, 1.0)



from typing import Protocol

from trajax.types.trajectories import JaxPathParameters, JaxPositions, JaxHeadings
from trajax.types.costs.common import PositionExtractor

from jaxtyping import Array as JaxArray, Float


class JaxPathParameterExtractor[StateBatchT](Protocol):
    def __call__(self, states: StateBatchT, /) -> JaxPathParameters:
        """Extracts path parameters from a batch of states."""
        ...


class JaxPathVelocityExtractor[InputBatchT](Protocol):
    def __call__(self, inputs: InputBatchT, /) -> Float[JaxArray, "T M"]:
        """Extracts path velocities from a batch of control inputs."""
        ...


class JaxPositionExtractor[StateBatchT](
    PositionExtractor[StateBatchT, JaxPositions], Protocol
): ...


class JaxHeadingExtractor[StateBatchT](Protocol):
    def __call__(self, states: StateBatchT, /) -> JaxHeadings:
        """Extracts heading angles from a batch of states."""
        ...



from typing import Self
from dataclasses import dataclass

from trajax.types import JaxObstacleSimulator
from trajax.obstacles.accelerated import JaxObstacleStatesForTimeStep

from numtypes import Array, Dims, D
from jaxtyping import Array as JaxArray, Float

import jax.numpy as jnp


@dataclass(frozen=True)
class JaxStaticObstacleSimulator[K: int](
    JaxObstacleSimulator[JaxObstacleStatesForTimeStep[K]]
):
    positions: Float[JaxArray, "K 2"]
    headings: Float[JaxArray, "K"]

    @staticmethod
    def empty() -> "JaxStaticObstacleSimulator[D[0]]":
        positions = jnp.empty((0, 2))

        return JaxStaticObstacleSimulator.create(positions=positions)

    @staticmethod
    def create[K_: int = int](
        *,
        positions: Array[Dims[K_, D[2]]] | Float[JaxArray, "K 2"],
        headings: Array[Dims[K_]] | Float[JaxArray, "K"] | None = None,
        obstacle_count: K_ | None = None,
    ) -> "JaxStaticObstacleSimulator[K_]":
        count = positions.shape[0]

        assert obstacle_count is None or obstacle_count == count, (
            f"Expected {obstacle_count} obstacles, but got {count}."
        )

        headings = headings if headings is not None else jnp.zeros(shape=(count,))

        return JaxStaticObstacleSimulator(
            positions=jnp.asarray(positions), headings=jnp.asarray(headings)
        )

    def with_time_step_size(self, time_step_size: float) -> Self:
        # Time step does not matter.
        return self

    def step(self) -> JaxObstacleStatesForTimeStep[K]:
        return JaxObstacleStatesForTimeStep.create(
            x=self.positions[:, 0], y=self.positions[:, 1], heading=self.headings
        )

    @property
    def obstacle_count(self) -> int:
        return self.positions.shape[0]



from .basic import (
    NumPyBicycleState as NumPyBicycleState,
    NumPyBicycleStateSequence as NumPyBicycleStateSequence,
    NumPyBicycleStateBatch as NumPyBicycleStateBatch,
    NumPyBicyclePositions as NumPyBicyclePositions,
    NumPyBicycleControlInputSequence as NumPyBicycleControlInputSequence,
    NumPyBicycleControlInputBatch as NumPyBicycleControlInputBatch,
    NumPyBicycleObstacleStateSequences as NumPyBicycleObstacleStateSequences,
    NumPyBicycleModel as NumPyBicycleModel,
    NumPyBicycleObstacleModel as NumPyBicycleObstacleModel,
)
from .accelerated import (
    JaxBicycleState as JaxBicycleState,
    JaxBicycleStateSequence as JaxBicycleStateSequence,
    JaxBicycleStateBatch as JaxBicycleStateBatch,
    JaxBicyclePositions as JaxBicyclePositions,
    JaxBicycleControlInputSequence as JaxBicycleControlInputSequence,
    JaxBicycleControlInputBatch as JaxBicycleControlInputBatch,
    JaxBicycleObstacleStateSequences as JaxBicycleObstacleStateSequences,
    JaxBicycleModel as JaxBicycleModel,
    JaxBicycleObstacleModel as JaxBicycleObstacleModel,
)



from typing import Any
from dataclasses import dataclass

from trajax.types import (
    StateSequence,
    SimulationData,
    Metric,
    Positions,
    PositionExtractor,
    Trajectory,
    LateralPositions,
)
from trajax.collectors import access

from numtypes import Array, Dims, shape_of

import numpy as np


@dataclass(kw_only=True, frozen=True)
class ComfortMetricResult[T: int = int]:
    lateral_acceleration: Array[Dims[T]]
    lateral_jerk: Array[Dims[T]]


@dataclass(kw_only=True, frozen=True)
class ComfortMetric[StateBatchT, PositionsT, LateralT](
    Metric[ComfortMetricResult[Any]]
):
    reference: Trajectory[Any, Any, PositionsT, LateralT]
    time_step_size: float
    position_extractor: PositionExtractor[StateBatchT, PositionsT]

    @staticmethod
    def create[S, P: Positions, L: LateralPositions](
        *,
        reference: Trajectory[Any, Any, P, L],
        time_step_size: float,
        position_extractor: PositionExtractor[S, P],
    ) -> "ComfortMetric[S, P, L]":
        assert time_step_size > 0, (
            f"Time step size must be positive, got {time_step_size}"
        )

        return ComfortMetric(
            reference=reference,
            time_step_size=time_step_size,
            position_extractor=position_extractor,
        )

    def compute[T: int = int](self, data: SimulationData) -> ComfortMetricResult[Any]:
        states = data(
            access.states.assume(StateSequence[T, Any, StateBatchT]).require()
        )
        state_batch = states.batched()

        positions = self.position_extractor(state_batch)
        lateral = self.reference.lateral(positions)

        lateral_array = np.asarray(lateral)[:, 0]
        lateral_velocity = checked_gradient(lateral_array, self.time_step_size)
        lateral_acceleration = checked_gradient(lateral_velocity, self.time_step_size)
        lateral_jerk = checked_gradient(lateral_acceleration, self.time_step_size)

        return ComfortMetricResult(
            lateral_acceleration=lateral_acceleration, lateral_jerk=lateral_jerk
        )

    @property
    def name(self) -> str:
        return "comfort"


def checked_gradient[L: int](y: Array[Dims[L]], dx: float) -> Array[Dims[L]]:
    if (points := y.shape[0]) < 2:
        return np.zeros_like(y)

    gradient = np.gradient(y, dx)

    assert shape_of(gradient, matches=(points,), name="gradient")

    return gradient



from dataclasses import dataclass

from trajax.types import (
    DataType,
    ObstacleIdAssignment,
    ObstacleStatesForTimeStep,
    ObstacleStatesHistory,
    JaxObstaclePositionsForTimeStep,
    JaxObstaclePositions,
    JaxObstaclePositionExtractor,
    NumPyObstaclePositionsForTimeStep,
    NumPyObstaclePositions,
    NumPyObstaclePositionExtractor,
)
from trajax.obstacles.history.accelerated import JaxObstacleIds
from trajax.obstacles.history.basic import NumPyObstacleIds
from trajax.obstacles.assignment.basic import NumPyHungarianObstacleIdAssignment

from numtypes import Array, Dims

import numpy as np
import jax.numpy as jnp


type JaxPositionsExtractor[StatesT, HistoryT] = JaxObstaclePositionExtractor[
    StatesT, HistoryT, JaxObstaclePositionsForTimeStep, JaxObstaclePositions
]


@dataclass(frozen=True)
class NumPyAdaptedObstaclePositionsForTimeStep[D_p: int, K: int](
    NumPyObstaclePositionsForTimeStep[D_p, K]
):
    _array: Array[Dims[D_p, K]]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[D_p, K]]:
        return self.array

    @property
    def dimension(self) -> D_p:
        return self.array.shape[0]

    @property
    def count(self) -> K:
        return self.array.shape[1]

    @property
    def array(self) -> Array[Dims[D_p, K]]:
        return self._array


@dataclass(frozen=True)
class NumPyAdaptedObstaclePositions[T: int, D_p: int, K: int](
    NumPyObstaclePositions[T, D_p, K]
):
    _array: Array[Dims[T, D_p, K]]

    def __array__(self, dtype: DataType | None = None) -> Array[Dims[T, D_p, K]]:
        return self.array

    @property
    def horizon(self) -> T:
        return self.array.shape[0]

    @property
    def dimension(self) -> D_p:
        return self.array.shape[1]

    @property
    def count(self) -> K:
        return self.array.shape[2]

    @property
    def array(self) -> Array[Dims[T, D_p, K]]:
        return self._array


@dataclass(kw_only=True)
class PositionExtractorAdapter[StatesT, HistoryT](
    NumPyObstaclePositionExtractor[
        StatesT, HistoryT, NumPyObstaclePositionsForTimeStep, NumPyObstaclePositions
    ]
):
    inner: JaxObstaclePositionExtractor[
        StatesT, HistoryT, JaxObstaclePositionsForTimeStep, JaxObstaclePositions
    ]

    @staticmethod
    def adapt(
        extractor: JaxPositionsExtractor[StatesT, HistoryT],
    ) -> "PositionExtractorAdapter[StatesT, HistoryT]":
        return PositionExtractorAdapter(inner=extractor)

    def of_states_for_time_step(
        self, states: StatesT, /
    ) -> NumPyObstaclePositionsForTimeStep:
        return NumPyAdaptedObstaclePositionsForTimeStep(
            np.asarray(self.inner.of_states_for_time_step(states))
        )

    def of_states(self, states: HistoryT, /) -> NumPyObstaclePositions:
        return NumPyAdaptedObstaclePositions(np.asarray(self.inner.of_states(states)))


@dataclass(frozen=True)
class JaxHungarianObstacleIdAssignment[
    StatesT: ObstacleStatesForTimeStep,
    HistoryT: ObstacleStatesHistory,
](ObstacleIdAssignment[StatesT, JaxObstacleIds, HistoryT]):
    # NOTE: Internally the Hungarian assignment is still done using SciPy and NumPy.
    inner: NumPyHungarianObstacleIdAssignment[StatesT, HistoryT]

    @staticmethod
    def create[S: ObstacleStatesForTimeStep, H: ObstacleStatesHistory](
        *,
        position_extractor: JaxPositionsExtractor[S, H],
        cutoff: float,
        start_id: int = 0,
    ) -> "JaxHungarianObstacleIdAssignment":
        return JaxHungarianObstacleIdAssignment(
            inner=NumPyHungarianObstacleIdAssignment.create(
                position_extractor=PositionExtractorAdapter.adapt(position_extractor),
                cutoff=cutoff,
                start_id=start_id,
            )
        )

    def __call__(
        self, states: StatesT, /, *, history: HistoryT, ids: JaxObstacleIds
    ) -> JaxObstacleIds:
        return JaxObstacleIds.create(
            ids=jnp.asarray(
                self.inner(
                    states, history=history, ids=NumPyObstacleIds(np.asarray(ids))
                )
            )
        )



from dataclasses import dataclass, field

from trajax.types import Risk, RiskMetric
from trajax.collectors.access import access
from trajax.collectors.registry import (
    DataTransformer,
    IdentityTransformer,
    ModificationNotifierMixin,
    ListCollectorMixin,
    OnModifyCallback,
)


@dataclass(frozen=True)
class RiskCollector[
    CostFunctionT,
    StateBatchT,
    ObstacleStatesT,
    SamplerT,
    RiskT = Risk,
](
    RiskMetric[CostFunctionT, StateBatchT, ObstacleStatesT, SamplerT, RiskT],
    ModificationNotifierMixin,
    ListCollectorMixin,
):
    inner: RiskMetric[CostFunctionT, StateBatchT, ObstacleStatesT, SamplerT, RiskT]
    transformer: DataTransformer[RiskT]
    _callbacks: list[OnModifyCallback] = field(default_factory=list)
    _collected: list[RiskT] = field(default_factory=list)

    @staticmethod
    def decorating[CF, SB, OS, S, R](
        metric: RiskMetric[CF, SB, OS, S, R],
        *,
        transformer: DataTransformer[R] = IdentityTransformer(),
    ) -> "RiskCollector[CF, SB, OS, S, R]":
        return RiskCollector(metric, transformer=transformer)

    def compute(
        self,
        cost_function: CostFunctionT,
        *,
        states: StateBatchT,
        obstacle_states: ObstacleStatesT,
        sampler: SamplerT,
    ) -> RiskT:
        self._collected.append(
            risk := self.inner.compute(
                cost_function,
                states=states,
                obstacle_states=obstacle_states,
                sampler=sampler,
            )
        )

        # TODO: Missing notify call.

        return risk

    @property
    def name(self) -> str:
        return self.inner.name

    @property
    def key(self) -> str:
        return access.risks.key
